/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *                      Simon Heath - MSKCC                                 *
 *                                                                          *
 *                          August 2000                                     *
 *                                                                          *
 * handle_res.c:                                                            *
 *                                                                          *
 * Routines for handling residual variance                                  *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <float.h>
#if HAVE_VALUES_H
#include <values.h>
#endif
#ifndef DBL_MAX
#define DBL_MAX MAXDOUBLE
#endif

#include "utils.h"
#include "loki.h"
#include "handle_res.h"

double res_prior_konst;

static void get_res_param(double *n,double *s)
{
	int i,j,type,nrec;
	double y;
  
	*s=*n=0.0;
	type=model.var.type;
	for(i=0;i<ped_size;i++) if(id_array[i].res) {
		if(type&ST_CONSTANT) nrec=1;
		else nrec=id_array[i].n_rec;
		for(j=0;j<nrec;j++) {
			y=id_array[i].res[j];
			if(!use_student_t) *s+=y*y;
			else *s+=y*y/id_array[i].vv[j];
		}
		*n+=(double)nrec;
	}
}

/* Sample residual variance */
double Sample_ResVar(void)
{
	double s,nn,y,v2;
  
	get_res_param(&nn,&s);
	nn+=RES_PRIOR_V0;
	s+=RES_PRIOR_V0*RES_PRIOR_S0;
	y=residual_var;
	residual_var=s/(sgamma(nn*0.5)*2.0);
	if(residual_var<residual_var_limit) residual_var=y;
	s/=nn;
	v2=nn*.5;
	y=log(v2)*v2-lgamma(v2)+log(s)*v2-log(residual_var)*(1.0+v2)-v2*s/residual_var;
	return y;
}

double Calc_Var_Prior(double v)
{
	return res_prior_konst-log(v)*(.5*RES_PRIOR_V0+1.0)-RES_PRIOR_S0*RES_PRIOR_V0/(2.0*v);
}

double Calc_ResLike(void)
{
	double s,nn;

	get_res_param(&nn,&s);
	return -.5*(nn*log(2.0*M_PI*residual_var)+s/residual_var);
}

double Calc_CensResLike(void)
{
	int i,j,type,idx;
	double s=0.0,nn=0.0,y,K=0.0,sd,z;

	sd=sqrt(residual_var*2.0);
	type=model.var.type;
	idx=model.var.var_index;
	if(type&ST_CONSTANT)	{
		for(i=0;i<ped_size;i++) if(id_array[i].res) {
			y=id_array[i].res[0];
			if(!censor_mode && (type&ST_CENSORED) && id_array[i].cens) {
				z=.5*erfc((y-id_array[i].cens[0])/sd);
				if(z>0.0) K+=log(z);
				else return -DBL_MAX;
			} else {
				s+=y*y;
				nn+=1.0;
			}
		}
	} else for(i=0;i<ped_size;i++) if(id_array[i].res)	{
		for(j=0;j<id_array[i].n_rec;j++)	{
			y=id_array[i].res[j];
			if(!censor_mode && id_array[i].cens && (id_array[i].data1[j][idx].flag&2))	{
				z=.5*erfc((y-id_array[i].cens[j])/sd);
				if(z>0.0) K+=log(z);
				else return -DBL_MAX;
			} else {
				s+=y*y;
				nn+=1.0;
			}
		}
	}
	return K-.5*(nn*log(2.0*M_PI*residual_var)+s/residual_var);
}

double Calc_Res_Ratio(double v1,double v2)
{
	double s,nn,l1;

	get_res_param(&nn,&s);
	l1=-.5*(nn*log(2.0*M_PI*v1)+s/v1);
	l1-=(-.5*(nn*log(2.0*M_PI*v2)+s/v2));
	return l1;
}

static double Calc_vprob(double v,double s,double x)
{
	v*=.5;
	return log(v)*v-lgamma(v)+log(s)*v-log(x)*(1.0+v)-v*s/x;
}

double Calc_Resprop(void)
{
	double s,v;

	get_res_param(&v,&s);
	v+=RES_PRIOR_V0;
	s+=RES_PRIOR_V0*RES_PRIOR_S0;
	s/=v;
	return Calc_vprob(v,s,residual_var);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Recalc_Res"
double Recalc_Res(void)
{
	int i,j,k,k1,k2,rec,nrec,type,mtype,idx;
	struct id_data *data;
	double x,y,er=0.0;
	
	mtype=model.var.type;
	idx=model.var.var_index;
	for(i=0;i<ped_size;i++) if(id_array[i].res) {
		nrec=id_array[i].n_rec;
		if(!nrec) nrec=1;
		for(rec=0;rec<nrec;rec++) {
			y= -grand_mean;
			if(mtype&ST_CONSTANT) {
				data=id_array[i].data+idx;
				if(data->flag&2) y+=id_array[i].cens[0];
			} else {
				data=id_array[i].data1[rec]+idx;
				if(data->flag&2) y+=id_array[i].cens[rec];
			}
			if(data->flag&ST_INTTYPE) y+=(double)data->data.value;
			else y+=data->data.rvalue;
			for(j=0;j<n_tloci;j++) if(tlocus[j].flag) {
				k2=tlocus[j].gt[i]-1;
				if(k2) y-=tlocus[j].eff[k2-1];
			}
			for(k=0;k<model.n_terms;k++) {
				type=model.term[k].vars[0].type;
				if(type&ST_TRAITLOCUS) continue;
				if(type&ST_ID) {
					y-=id_array[i].bv;
					continue;
				}
				k1=model.term[k].vars[0].var_index;
				if(type&ST_MARKER) {
					k2=marker[k1].gt[i]-1;
					if(k2) y-=model.term[k].eff[k2-1];
				} else {
					data=0;
					if(type&ST_CONSTANT) {
						if(id_array[i].data) data=id_array[i].data+k1;
					} else if(id_array[i].data1) data=id_array[i].data1[rec]+k1;
					if(!data) ABT_FUNC("Internal error - null data pointer\n");
					if(type&ST_FACTOR) {
						k2=(int)data->data.value-1;
						if(!(type&ST_RANDOM)) k2--;
						if(k2>=0) y-=model.term[k].eff[k2];
					} else {
						if(data->flag&ST_INTTYPE) x=(double)data->data.value;
						else x=data->data.rvalue;
						y-=model.term[k].eff[0]*x;
					}
				}
			}
			er+=fabs(id_array[i].res[rec]-y);
			id_array[i].res[rec]=y;
		}
	}
	return er;
}
