/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                         - Rockefeller University                         *
 *                                                                          *
 *                        August 1997                                       *
 *                                                                          *
 * loki_peel.c:                                                             *
 *                                                                          *
 * Perform peeling calculations                                             *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <float.h>
#if HAVE_VALUES_H
#include <values.h>
#endif
#ifndef DBL_MAX
#define DBL_MAX MAXDOUBLE
#endif

#include "ranlib.h"
#include "utils.h"
#include "loki.h"
#include "loki_peel.h"

static double *qval,*pval,*mval,**freq,*peel_famval,**complex_freq,**complex_pen,*complex_out_p;
static struct R_Func *rf;
static struct Peelseq_Head **peel_list;
static unsigned long **a_set,mask,*peel_pt1;
static int n_idx,n_all,linktype,max_fnd,max_all,out_p_size;
static int n_bits,n_terms,hash_mode,*peel_alls,*complex_mem,max_pen,max_pen1;
static size_t hb_size=2048,mb_size=4096;
static struct fset *peel_fs;
int peel_trace;

static struct hash_element *hashtable[HASHTABLE_SIZE];
static struct hash_block *first_hash_block,*hash_block;

#define MRK_MBLOCK 0
#define TRT_MBLOCK 1
static struct peel_mem_block *first_mem_block[2],*mem_block[2];

static lk_ulong old_index;
static struct hash_element *old_element;

int si_mode;

void print_allele_type1(FILE *fptr,const int locus,const int j)
{
	if(j==marker[locus].lumped) fputs(LUMPED_ALLELE,fptr);
	else if(marker[locus].rec_flag==ST_STRING) fputs(marker[locus].recode[j].string,fptr);
	else fprintf(fptr,"%d",marker[locus].recode[j].value);
}

/* Print the original allele code for (recoded) allele i */
void print_allele_type(FILE *fptr,const int locus,const int comp,const int i)
{
	print_allele_type1(fptr,locus,allele_trans[locus][comp][i]);
}

/* Print original code for maternal or paternal allele (depending on flag) for
 * individual id.  Takes account of set recoding */
void print_allele_name(FILE *fptr,const int id,const int locus,const int flag)
{
	unsigned long c;
	int i,j,allele,comp;
	
	comp=id_array[id].comp;
	allele=id_array[id].allele[flag]-1;
	c=1<<allele;
	if(c&req_set[flag][locus][id]) {
		i=j=0;
		c=req_set[flag][locus][id];
		while(c)	{
			if(c&1) {
				fputc(j++?',':'[',fptr);
				print_allele_type(fptr,locus,comp,i);
			}
			i++;
			c>>=1;
		}
		fputc(']',fptr);
	} else print_allele_type(fptr,locus,comp,allele);
}

/* Given index x (from R-Function), returns corresponding n allele types in gt */
static void get_gts(lk_ulong x,const int n,int *gt)
{
	int i;
	lk_ulong a;
	
	a=(1<<n_bits)-1;
	i=0;
	while(x)	{
		gt[i++]=1+(int)(x&a);
		x>>=n_bits;
	}
	for(;i<n;i++) gt[i]=1;
}

/* The inverse of get_gts() - converts the n allele types in gt into an index */
static lk_ulong get_index1(int n,int *gt)
{
	int i;
	lk_ulong x;
	
	x=0;
	for(i=n-1;i>=0;i--) {
		x<<=n_bits;
		x|=gt[i]-1;
	}
	return x;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_new_element"
/* Returns storage for storing a new non-zero element.  Elements are allocated
 * in blocks of size hb_size */
static struct hash_element *get_new_element(void)
{
	struct hash_element *element;
	
	while(hash_block->ptr>=hash_block->size) {
		if(!hash_block->next)
		  if(!(hash_block->next=calloc(1,sizeof(struct hash_block)))) ABT_FUNC(MMsg);
		hash_block=hash_block->next;
		if(!hash_block->elements) {
			if(!(hash_block->elements=malloc(sizeof(struct hash_element)*hb_size))) ABT_FUNC(MMsg);
			hash_block->size=hb_size;
			hash_block->ptr=0;
		}
	}
	element=hash_block->elements+hash_block->ptr++;
	element->left=element->right=0;
	element->index=0;
	element->p=0.0;
	n_terms++;
	return element;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_new_memblock"
static struct peel_mem_block *get_new_memblock(size_t size,int flag)
{
	struct peel_mem_block *p;
	
	if(!(p=malloc(sizeof(struct peel_mem_block)))) ABT_FUNC(MMsg);
	p->index=0;
	p->val=0;
	p->size=size;
	p->ptr=0;
	p->next=0;
	if(flag==MRK_MBLOCK) if(!(p->index=malloc(sizeof(lk_ulong)*size))) ABT_FUNC(MMsg);
	if(!(p->val=malloc(sizeof(double)*size))) ABT_FUNC(MMsg);
	return p;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_rf_memory"
static void get_rf_memory(struct R_Func *rf,size_t size,int flag)
{
	size_t i;
	struct peel_mem_block *p,*p1;
	
	p1=mem_block[flag];
	i=p1->size-p1->ptr;
	if(size>i) {
		p=p1;
		while(p->next)	{
			if(size<=p->next->size) break;
			p=p->next;
		}
		if(!(p->next)) {
			hb_size*=1.2;
			i=(size>hb_size)?size:hb_size;
			p=get_new_memblock(i,flag);
		} else {
			p1=p->next;
			p->next=p1->next;
			p=p1;
			p->ptr=0;
			p1=mem_block[flag];
		}
		p->next=p1->next;
		p1->next=p;
		p1=p;
	}
	i=p1->ptr;
	rf->index=p1->index+i;
	rf->p=p1->val+i;
	p1->ptr+=size;
	mem_block[flag]=p1;
}

/* Insert element with index idx into the binary tree hanging off *node */
static struct hash_element *insert_node(struct hash_element *node,lk_ulong idx)
{
	if(idx<node->index) {
		if(node->left) return insert_node(node->left,idx);
		return node->left=get_new_element();
	} else if(idx>node->index)	{
		if(node->right) return insert_node(node->right,idx);
		return node->right=get_new_element();
	}
	return node;
}

/* Recursive routine for getting all elements from binary tree hanging off *node and storing
 * the values and indices from them in *tp and *tl */
static void get_nodes(struct hash_element *node,double *tp,lk_ulong *tl,int *j)
{
	if(node->left)	get_nodes(node->left,tp,tl,j);
	tp[*j]=node->p;
	tl[(*j)++]=node->index;
	if(node->right) get_nodes(node->right,tp,tl,j);
}

/* Similar to get_nodes, but stops when the cumulative probability >= z.  Allows sampling
 * from distribution */
static int sample_comb(struct hash_element *node,const double z,double *p,lk_ulong *idx)
{
	if(node->left) if(sample_comb(node->left,z,p,idx)) return 1;
	if(node->p>0.0) {
		*p+=node->p;
		if(*p>=z) {
			*idx=node->index;
			return 1;
		}
	}
	if(node->right) if(sample_comb(node->right,z,p,idx)) return 1;
	return 0;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "add_to_term"
/* Adds p to output element defined by the n_out allele types in gt_store.  If the element
 * does not exist, it is created. */
static void add_to_term(const int n_out,int *gt_store,const double p)
{
	lk_ulong idx;
	int i;
	struct hash_element *element;

	idx=get_index1(n_out,gt_store);
	if(hash_mode) {/* If all possible combinations can fit into hashtable, we don't need to search */
		i=(int)idx;
		if(!hashtable[i])	{
			hashtable[i]=get_new_element();
			hashtable[i]->p=p;
		} else hashtable[i]->p+=p;
	} else { /* ...else we need to search */
		if(old_element) if(idx==old_index) { /* Avoid search if the index is the same as the last one */
			old_element->p+=p;
			return;
		}
		i=(int)(idx%HASHTABLE_SIZE); /* Get hash index */
		if(hashtable[i]) element=insert_node(hashtable[i],idx);
		else element=hashtable[i]=get_new_element();
		element->p+=p;
		element->index=idx;
		old_index=idx;
		old_element=element;
	}
}

/* Clean up memory used */
static void free_hash_blocks(void)
{
	struct hash_block *hb1;
	
	hash_block=first_hash_block;
	while(hash_block) {
		if(hash_block->elements) free(hash_block->elements);
		hb1=hash_block->next;
		free(hash_block);
		hash_block=hb1;
	}
	first_hash_block=0;
}

void free_complex_mem(void)
{
	if(complex_mem) free(complex_mem);
	if(complex_out_p) free(complex_out_p);
	if(complex_freq) {
		if(complex_freq[0]) free(complex_freq[0]);
		free(complex_freq);
	}
	if(complex_pen) {
		if(complex_pen[0]) free(complex_pen[0]);
		free(complex_pen);
	}
	free_hash_blocks();
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "setup_complex_peel"
static void setup_complex_peel(const struct Complex_Element *element, const int sampling,int *temp[],int *n_other)
{
	int i,j,k,k1,n_peel,n_inv,n_rf,*inv;
	static int complex_mem_size;
	
	n_inv=element->n_involved;
	n_peel=element->n_peel;
	n_rf=element->n_rfuncs;
	inv=element->involved;
	i=7*n_inv+n_rf+2*n_peel;
	if(!complex_mem) {
		complex_mem_size=i;
		if(!(complex_mem=calloc((size_t)i,sizeof(int)))) ABT_FUNC(MMsg);
	} else {
		if(i>complex_mem_size) {
			complex_mem_size=i;
			if(!(complex_mem=realloc(complex_mem,i*sizeof(int)))) ABT_FUNC(MMsg);
		}
		memset(complex_mem,0,i*sizeof(int));
	}
	temp[0]=complex_mem;
	for(i=1;i<8;i++) temp[i]=temp[i-1]+n_inv;
	for(;i<10;i++) temp[i]=temp[i-1]+n_peel;
	/* 'other' alleles are alleles not in R-Functions and not already sampled
	 * (if on reverse sampling pass) */
	k1=sampling?n_peel:n_inv; /* When we are sampling, the output alleles have already been sampled */
	for(*n_other=i=0;i<k1;i++) if(!(element->flags[i]&IN_RF)) {
		if(element->flags[i]&(HAP_JNT|HAP_DAT)) {
			k= -inv[i];
			for(j=0;j<n_inv;j++) if(inv[j]==k) temp[3][*n_other]=j;
		}
		temp[1][(*n_other)++]=i;
	}
	/* Make a list of alleles that we need transmission probs for */
	for(i=0;i<n_inv;i++) {
		if(element->flags[i]&HAD_P) {
			k=id_array[-inv[i]-1].sire;
			for(j=0;j<n_inv;j++) {
				if(inv[j]==k) temp[5+X_MAT][i]=j;
				else if(-inv[j]==k) temp[5+X_PAT][i]=j;
			}
		}
		if(element->flags[i]&HAD_M) {
			k=id_array[inv[i]-1].dam;
			for(j=0;j<n_inv;j++) {
				if(inv[j]==k) temp[5+X_MAT][i]=j;
				else if(-inv[j]==k) temp[5+X_PAT][i]=j;
			}
		}
	}
	/* Form list of alleles not in R-Functions */
	if(sampling) {
		for(i=0;i<n_peel;i++) temp[0][i]=0;
		for(;i<n_inv;i++) temp[0][i]=1;
	} else for(i=0;i<n_inv;i++) temp[0][i]=0;	
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_complex_peelop"
/* Perform general peeling operation defined in element (i.e., as opposed to a 'simple' 
 * nuclear family based peeling operation) */
static double loki_complex_peelop(const struct Complex_Element *element,const int locus,const int s_flag,pen_func pen)
{
	int i,j,k,k1,k2,k3,k4,n_out,n_peel,n_inv,*inv,n_rf,n_ind,n_other,ef,ef1,sampling=0;
	int *gt_store,*gt_store1,*other_ptr,*other_list,*rf_ptr,*jnt_list,ht_size;
	int *off_index[2],*fnd_list,n_fnd=0,all,*pen_list,n_pen,*temp_p[10];
	double max_terms,fill,*tp,prob=0.0,z,p1,Konst=0.0;
	lk_ulong a,b,m,*tl;
	struct hash_element *elem;
	
	/* Get details about peeling operation */
	n_inv=element->n_involved; /* No. alleles involved in op */
	n_peel=element->n_peel; /* No. alleles to peel out (absorb) */
	n_out=n_inv-n_peel; /* No. to appear in output R-Function */
	n_rf=element->n_rfuncs; /* No. input R-Functions */
	inv=element->involved; /* List of involved alleles (alleles to peel come first) */
	/* Allocate first hash_block, if not already done so */
	if(!first_hash_block) {
		if(!(first_hash_block=malloc(sizeof(struct hash_block)))) ABT_FUNC(MMsg);
		if(!(first_hash_block->elements=malloc(sizeof(struct hash_element)*hb_size))) ABT_FUNC(MMsg);
		first_hash_block->next=0;
		first_hash_block->size=hb_size;
	}
	/* Reset pointers to all hash_blocks */
	hash_block=first_hash_block;
	while(hash_block) {
		hash_block->ptr=0;
		hash_block=hash_block->next;
	}
	hash_block=first_hash_block;
	n_terms=0; /* No. non-zero terms in output R-Function */
	/* if s_flag is non-zero then we are doing a sampling run */
	/* if s_flag&OP_SAMPLING then we are on the reverse (sampling) phase */
	/* In any case, if !n_out and s_flag then we can sample */
	if(s_flag) {
		if(s_flag&OP_SAMPLING) sampling=1;
		else if(!n_out) sampling=1;
	}
	/* Check size of function we are assembling.  Check index for output function will
	 * fit into a lk_long.  Compute max_terms, the number of output terms if they are all non-zero.
	 * See if we need to do hash search, or if output will fit into hashtable */
	/* If sampling then we assemble function on the peeled alleles */
	if(sampling) {
		if(n_bits*n_peel>(int)LK_LONG_BIT) {
			fprintf(stderr,"\nToo many individuals in output R-Function for marker %s\nn_peel = %d, n_all = %d, n_bits = %d, required size = %d, LONG_BIT = %d\n",marker[locus].name,n_peel,n_all,n_bits,n_peel*n_bits,LK_LONG_BIT);
			ABT_FUNC("Aborting\n");
		}
		max_terms=log((double)n_all)*(double)n_peel;
		hash_mode=(log(2.0)*n_bits*n_peel<log((double)HASHTABLE_SIZE));
		ht_size=hash_mode?1<<(n_bits*n_peel):HASHTABLE_SIZE;
	} else { /* Otherwise we assemble function on output alleles */
		if(n_bits*n_out>(int)LK_LONG_BIT) {
			fprintf(stderr,"\nToo many individuals in output R-Function for marker %s\nn_out = %d, n_all = %d, n_bits = %d, required size = %d, LONG_BIT = %d\n",marker[locus].name,n_out,n_all,n_bits,n_out*n_bits,LK_LONG_BIT);
			ABT_FUNC("Aborting\n");
		}
		max_terms=log((double)n_all)*(double)n_out;
		hash_mode=(log(2.0)*n_bits*n_out<log((double)HASHTABLE_SIZE));
		ht_size=hash_mode?1<<(n_bits*n_out):HASHTABLE_SIZE;
	}
	old_element=0;
	for(i=0;i<ht_size;i++) hashtable[i]=0;
	setup_complex_peel(element,sampling,temp_p,&n_other);
	gt_store=temp_p[0];
	other_list=temp_p[1];
	other_ptr=temp_p[2];
	jnt_list=temp_p[3];
	gt_store1=temp_p[4];
	off_index[0]=temp_p[5];
	off_index[1]=temp_p[6];
	pen_list=temp_p[7];
	fnd_list=temp_p[8];
	rf_ptr=temp_p[9];
	/* Compute masks (first run through only) - used for finding mutually 
	 * consistent terms from multiple input R-Functions */
 	if(n_rf && !(marker[locus].flag&RFMASK_OK)) {
		for(i=0;i<n_rf;i++) {
			j=element->index[i];
			n_ind=rf[j].n_ind;
			a=0;
			k2=0;
			b=(1<<n_bits)-1;
			for(k=n_ind-1;k>=0;k--) {
				a<<=n_bits;
				k1=rf[j].id_list[k];
				if(gt_store[k1]) a|=b;
				else {
					k2|=1<<k1;
					gt_store[k1]=1;
				}
			}
			rf[j].mask[sampling]=a;
			rf[j].mask1[sampling]=k2;
	  	}
	}
	/* Make a list of alleles we need penetrances and/or founder probs for */
	n_pen=0;
	if(n_peel) {
		/* Make list of founder alleles being peeled */
		for(k=0;k<n_peel;k++) {
			j=abs(inv[k])-1;
			k1=id_array[j].sire;
			if(k1 && !id_array[k1-1].pruned_flag[locus]) continue;
			n_fnd++;
		}
		/* Get frequency info. for founder alleles (tricky because of recoding) */
		if(n_fnd) {
			if(!complex_freq) {
				if(!(complex_freq=malloc(sizeof(void *)*n_fnd))) ABT_FUNC(MMsg);
				max_all=n_all*n_fnd;
				if(!(complex_freq[0]=malloc(sizeof(double)*max_all))) ABT_FUNC(MMsg);
				max_fnd=n_fnd;
			} else {
				if(n_fnd>max_fnd) {
					max_fnd=n_fnd;
					tp=complex_freq[0];
					if(!(complex_freq=realloc(complex_freq,sizeof(void *)*n_fnd))) ABT_FUNC(MMsg);
					complex_freq[0]=tp;
				}
				if(n_fnd*n_all>max_all) {
					max_all=n_all*n_fnd;
					if(!(complex_freq[0]=realloc(complex_freq[0],sizeof(double)*max_all))) ABT_FUNC(MMsg);
				}
			}
			for(i=1;i<n_fnd;i++) complex_freq[i]=complex_freq[i-1]+n_all;
			n_fnd=0;
			for(k=0;k<n_peel;k++) {
				j=abs(inv[k])-1;
				k1=id_array[j].sire;
				if(k1 && !id_array[k1-1].pruned_flag[locus]) continue;
				k3=id_array[j].group-1;
				if(k3<0) ABT_FUNC("Internal error - and group number\n");
				if(inv[k]<0) {
					a=req_set[X_PAT][locus][j];
					k2=id_array[j].lumped[X_PAT];
				} else {
					a=req_set[X_MAT][locus][j];	
					k2=id_array[j].lumped[X_MAT];
				}
				complex_freq[n_fnd][k2]=0.0;
				for(k1=0;k1<n_all;k1++) {
					if(a&(1<<k1)) complex_freq[n_fnd][k2]+=freq[k3][k1];
					else complex_freq[n_fnd][k1]=freq[k3][k1];
				}
				fnd_list[n_fnd++]=k;
			}
		}
		if(pen) {
			/* Find alleles which we need penetrances for */
			for(k=0;k<n_peel;k++) {
				k1=abs(inv[k])-1;
				if(linktype==LINK_AUTO || (linktype==LINK_X && id_array[k1].sex==2)) {
					for(k2=k+1;k2<n_inv;k2++) if(inv[k2]== -inv[k]) break;
					if(k2==n_inv) continue;
					if(id_array[k1].res) pen_list[n_pen++]=k1;
				} else if(linktype==LINK_X && inv[k]>0 && id_array[k1].res) pen_list[n_pen++]=k1;
				else if(linktype==LINK_X && inv[k]<0 && id_array[k1].res) pen_list[n_pen++]=k1;
			}
			/* Pre-calculate penetrances */
			if(n_pen) {
				if(!complex_pen) {
					if(!(complex_pen=malloc(sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
					max_pen1=n_pen*n_idx;
					if(!(complex_pen[0]=malloc(sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
					max_pen=n_pen;
				} else {
					if(n_pen>max_pen)	{
						max_pen=n_pen;
						if(!(complex_pen=realloc(complex_pen,sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
					}
					if(n_pen*n_idx>max_pen1) {
						max_pen1=n_pen*n_idx;
						if(!(complex_pen[0]=realloc(complex_pen[0],sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
					}
				}
				for(k1=0;k1<n_pen*n_idx;k1++) complex_pen[0][k1]=1.0;
				for(k1=1;k1<n_pen;k1++) complex_pen[k1]=complex_pen[k1-1]+n_idx;
				for(k=0;k<n_pen;k++)	{
					pen(complex_pen[k],pen_list[k],locus,n_all,n_bits);
					z=0.0;
					for(k1=0;k1<n_idx;k1++) z+=complex_pen[k][k1];
					for(k1=0;k1<n_idx;k1++) complex_pen[k][k1]/=z;
					Konst+=log(z);
				}
			}
		}
	}
	for(i=0;i<n_inv;i++) gt_store[i]=0;
	ef1=0;
	i=0;
	/* Main loop.
	 *
	 * (1) If sampling then fix allele types of already sampled alleles.
	 * (2) Find mutually conistent configurations from input R-Functions
	 * (3) Find consistent configurations for 'other' alleles
	 * (4) Compute transmission/founder/penetrance probs
	 * (5) If still possible then add to output function
	 * (6) Cycle until tired
	 */
	for(;;) {
		/* Fix allele types of sampled alleles */
		if(sampling) for(j=0;j<n_out;j++) {
			k=inv[j+n_peel];
			if(k>0) gt_store[j+n_peel]=id_array[k-1].allele[X_MAT];
			else gt_store[j+n_peel]=id_array[-k-1].allele[X_PAT];
		}
		/* Cycle through R-Functions */
		for(;i<n_rf;i++) {
			j=element->index[i];
			ef=0;
			n_ind=rf[j].n_ind;
			/* Find term consistent with what's already been set up */
			a=rf[j].mask[sampling];
			if(a)	{
				b=0;
				for(k=n_ind-1;k>=0;k--)	{
					b<<=n_bits;
					k1=rf[j].id_list[k];
					if(gt_store[k1]) b|=gt_store[k1]-1;
				}
				b&=a;
				k1=rf[j].n_terms;
				tl=rf[j].index;
				for(k=rf_ptr[i];k<k1;k++) if((tl[k]&a)==b) break;
				rf_ptr[i]=k;
			}
			/* If consistent term found then get corresponding allele types */
			if(rf_ptr[i]<rf[j].n_terms) {
				get_gts(rf[j].index[rf_ptr[i]],n_ind,gt_store1);
				for(k=0;k<n_ind;k++)	{
					k1=rf[j].id_list[k];
					gt_store[k1]=gt_store1[k];
				}
			} else { /* Inconsistency found - try new combination */
				k2=rf[j].mask1[sampling];
				k=0;
				while(k2) {
					if(k2&1) gt_store[k]=0;
					k++;
					k2>>=1;
				}
				do {
					if(!i) {
						ef1=1;
						break;
					}
					rf_ptr[i--]=0;
					rf_ptr[i]++;
					j=element->index[i];
					k2=rf[j].mask1[sampling];
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
				} while(rf_ptr[i]>=rf[j].n_terms);
				i--;
			}
			if(ef1) break;
		}
		if(ef1) break;
		/* We have a consistent configuration from the R-Functions, now find 1 for the 'other' alleles */
		for(ef=i=0;i<=n_other;i++)	{
			if(i==n_other)	{
				if(!ef) {
					for(k=0;k<n_inv;k++)	{
						if(inv[k]>0) {
							if(element->flags[k]&(HAP_JNT|HAP_DAT))	{
								j=inv[k]-1;
								for(k1=0;k1<n_inv;k1++) if(inv[k1]== -inv[k]) {
									if(!(a_set[j][gt_store[k]-1]&(1<<(gt_store[k1]-1)))) ef=1;
									break;
								}
							}
						}
						if(ef) break;
					}
					/* Add Parent-off transmission probs. */
					p1=1.0;
					if(!ef) for(k=0;k<n_inv;k++) {
						if(element->flags[k]&HAD_P) { /* Sire-offspring */
							j= -inv[k]-1;
							tp=id_array[j].tpp[X_PAT];
							m=1<<(gt_store[k]-1);
							a=req_set[X_PAT][locus][j];
							if(a&m) m=a;
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(m&(1<<(all-1))) z=tp[X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(m&(1<<(all-1))) z+=tp[X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
						} else if(element->flags[k]&HAD_M) { /* Dam-offspring */
							j=inv[k]-1;
							tp=id_array[j].tpp[X_MAT];
							m=1<<(gt_store[k]-1);
							a=req_set[X_MAT][locus][j];
							if(a&m) m=a;
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(m&(1<<(all-1))) z=tp[X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(m&(1<<(all-1))) z+=tp[X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
							if(element->flags[k]&(HAP_JNT|HAP_DAT)) {
								for(k1=0;k1<n_inv;k1++) if(inv[k1]== -inv[k]) {	
									j=inv[k]-1;
									if(!(a_set[j][gt_store[k]-1]&(1<<(gt_store[k1]-1)))) ef=1;
									break;
								}
								if(ef) break;
							}
						}
					}
					/* Add penetrances and founder probs and store result */
					if(!ef) {
						for(k=0;k<n_rf;k++) p1*=rf[element->index[k]].p[rf_ptr[k]];
						for(k1=0;k1<n_fnd;k1++) { /* Add founder frequencies */
							k=fnd_list[k1];
							p1*=complex_freq[k1][gt_store[k]-1];
						}
						/* Add penetrances for those being peeled if required */
 						if(n_pen) {
							for(k=0;k<n_inv;k++)	{
								k1=inv[k];
								if(k1<0) id_array[-k1-1].allele[X_PAT]=gt_store[k];
								else id_array[k1-1].allele[X_MAT]=gt_store[k];
							}
							for(k=0;k<n_pen;k++)	{
								k1=pen_list[k];
								k2=((id_array[k1].allele[X_PAT]-1)<<n_bits)|(id_array[k1].allele[X_MAT]-1);
								p1*=complex_pen[k][k2];
							}
						}
						prob+=p1;
						if(sampling) add_to_term(n_peel,gt_store,p1);
						else add_to_term(n_out,gt_store+n_peel,p1);
						ef=1;
					}
				}
			}
			/* Find next combination */
			if(ef) {
				i--;
				while(i>=0) {
					other_ptr[i]++;
					gt_store[other_list[i]]=0;
					if(other_ptr[i]<n_all) break;
					other_ptr[i--]=0;
				}
				i--;
				if(i<-1) break;
				ef=0;
				continue;
			}
			k1=other_list[i];
			if(gt_store[k1]) continue;
			j=abs(inv[k1])-1;
			if(element->flags[k1]&(HAP_JNT|HAP_DAT)) {
				k2=jnt_list[i];
				if(inv[k1]>0) {
					if(gt_store[k2]) {
						b=1<<(gt_store[k2]-1);
						for(k=other_ptr[i];k<n_all;k++) if(a_set[j][k]&b) {
							gt_store[k1]=k+1;
							break;
						}
						if((other_ptr[i]=k)==n_all) ef=1;
					} else {
						for(k=0;k<n_other;k++) if(other_list[k]==k2) break;
						if(k==n_other) ABT_FUNC("Internal error - no other?\n");
						if(i<k) {
							for(k3=other_ptr[i];k3<n_all;k3++) {
								a=a_set[j][k3];
								if(!a) continue;
								for(k4=other_ptr[k];k4<n_all;k4++) if(a&(1<<k4)) {
									gt_store[k1]=k3+1;
									gt_store[k2]=k4+1;
									break;
								}
								if(k4==n_all) other_ptr[k]=0;
								else {
									other_ptr[k]=k4;
									break;
								}
							}
							if((other_ptr[i]=k3)==n_all) ef=1;
						} else {
							for(k4=other_ptr[k];k4<n_all;k4++) {
								a=1<<k4;
								if(id_array[j].temp[X_PAT]&a)	{
									for(k3=other_ptr[i];k3<n_all;k3++) if(a_set[j][k3]&a) {
										gt_store[k1]=k3+1;
										gt_store[k2]=k4+1;
										break;
									}
									if(k3==n_all) other_ptr[i]=0;
									else {
										other_ptr[i]=k3;
										break;
									}
								} else other_ptr[i]=0;
							}
							if((other_ptr[k]=k4)==n_all) ef=1;
						}
					}
				} else {
					if(gt_store[k2]) {
						b=a_set[j][gt_store[k2]-1];
						for(k=other_ptr[i];k<n_all;k++) if(b&(1<<k))	{
							gt_store[k1]=k+1;
							break;
						}
						if((other_ptr[i]=k)==n_all) ef=1;
					} else {
						for(k=i+1;k<n_other;k++) if(other_list[k]==k2) break;
						if(k==n_other) ABT_FUNC("Internal error - why for this happen?\n");
						if(i<k) {
							for(k3=other_ptr[i];k3<n_all;k3++) {
								a=1<<k3;
								if(id_array[j].temp[X_PAT]&a)	{
									for(k4=other_ptr[k];k4<n_all;k4++) if(a_set[j][k4]&a)	{
										gt_store[k1]=k3+1;
										gt_store[k2]=k4+1;
										break;
									}
									if(k4==n_all) other_ptr[k]=0;
									else {
										other_ptr[k]=k4;
										break;
									}
								} else other_ptr[k]=0;
							}
							if((other_ptr[i]=k3)==n_all) ef=1;
						} else {
							for(k4=other_ptr[k];k4<n_all;k4++) {
								a=a_set[j][k4];
								if(!a) continue;
								for(k3=other_ptr[i];k3<n_all;k3++) if(a&(1<<k3)) {
									gt_store[k1]=k3+1;
									gt_store[k2]=k4+1;
									break;
								}
								if(k3==n_all) other_ptr[i]=0;
								else {
									other_ptr[i]=k3;
									break;
								}
							}
							if((other_ptr[k]=k4)==n_all) ef=1;
						}
					}
				}
			} else {
				if(inv[k1]>0) {
					for(k=other_ptr[i];k<n_all;k++) if(a_set[j][k]) {
						gt_store[k1]=k+1;
						break;
					}
					if((other_ptr[i]=k)==n_all) ef=1;
				} else {
					a=id_array[j].temp[X_PAT];
					for(k=other_ptr[i];k<n_all;k++) if(a&(1<<k)) {
						gt_store[k1]=k+1;
						break;
					}
					if((other_ptr[i]=k)==n_all) ef=1;
				}
			}
		}
		for(i=0;i<n_other;i++) gt_store[other_list[i]]=0;
		/* Cycle through R-Functions */
		if(n_rf)	{
			i=n_rf-1;
			do {
				rf_ptr[i]++;
				j=element->index[i];
				k2=rf[j].mask1[sampling];
				if(k2) {
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(rf_ptr[i]<rf[j].n_terms) break;
				}
				rf_ptr[i--]=0;
			} while(i>=0);
			if(i>=0) ef=0;
		} else i=0;
		if(ef) break;
	}
	/* All valid combinations have been visited.  Have we found any ? */
	if(!n_terms) {
		ABT_FUNC("Zero probability!\n");
	}
	/* If sampling then sample from output function */
	if(sampling) {
		do {
			z=ranf()*prob;
			p1=0.0;
			for(j=k=0;k<ht_size;k++) {
				elem=hashtable[k];
				if(!elem) continue;
				if(hash_mode) {
					if(elem->p>0.0) {
						p1+=elem->p;
						if(z<=p1) {
							a=(lk_ulong)k;
							break;
						}
					}
				} else if(sample_comb(elem,z,&p1,&a)) break;
			}
		} while(k==ht_size);
		get_gts(a,n_peel,gt_store);
		for(i=0;i<n_peel;i++) {
			j=inv[i];
			if(j>0) {
				id_array[j-1].allele[X_MAT]=gt_store[i];
				id_array[j-1].flag|=SAMPLED_MAT;
			} else {
				id_array[-j-1].allele[X_PAT]=gt_store[i];
				id_array[-j-1].flag|=SAMPLED_PAT;
			}
		}
	} else { /* Otherwise normalize and store */
		i=element->out_index;
		if(i>=0)	{
			rf[i].n_terms=n_terms;
			get_rf_memory(rf+i,n_terms,MRK_MBLOCK);
			tl=rf[i].index;
			tp=rf[i].p;
			for(j=k=0;k<ht_size;k++) {
				elem=hashtable[k];
				if(!elem) continue;
				if(hash_mode) {
					tp[j]=elem->p;
					tl[j++]=(lk_ulong)k;
				} else get_nodes(elem,tp,tl,&j);
			}
			fill=exp(log((double)n_terms)-max_terms);
			rf[i].fill=fill;
			for(j=0;j<n_terms;j++) tp[j]/=prob;
		} else fill=1.0;
	}
	return log(prob)+Konst;
}

/* Get parental probability dist., put into *val */
static void get_par_probs(double *val,const int i,const int locus,pen_func pen)
{
	int j,k,fflag=0;
	double f,f1,lf[2],*fq=0;
	unsigned long cm[2],a;
	lk_ulong l,l1,l2;
	
	j=id_array[i].sire;
	if(j && !(id_array[j-1].pruned_flag[locus])) fflag=1;
	else {
		k=id_array[i].group-1;
		if(k<0) ABT_FUNC("Internal error - bad group number\n");
		fq=freq[k];
	}
	/* If found add founder probs */
	if(!fflag) for(j=0;j<2;j++) {
		lf[j]=0.0;
		cm[j]=req_set[j][locus][i];
		if(cm[j]) {
			a=cm[j];
			for(k=0;k<n_all;k++) if(a&(1<<k)) lf[j]+=fq[k];
		}
	}
	for(k=0;k<n_idx;k++) val[k]=0.0;
	if((k=id_array[i].rfp)>=0) { /* Insert Previously computed R_Func */
		if(fflag) {
			for(j=0;j<rf[k].n_terms;j++) val[rf[k].index[j]]=rf[k].p[j];
		} else {
			for(j=0;j<rf[k].n_terms;j++) {
				l=rf[k].index[j];
				l1=l&mask;
				l2=(l>>n_bits)&mask;
				if((cm[X_MAT])&(1<<l1)) f=lf[X_MAT];
				else f=fq[l1];
				if((cm[X_PAT])&(1<<l2)) f1=lf[X_PAT];
				else f1=fq[l2];
				val[l]=rf[k].p[j]*f*f1;
			}
		}
	} else { /* No R-Function - remove illegal genotypes */
		if(fflag) {
			for(j=0;j<n_all;j++) {
				a=a_set[i][j];
				if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
					l=(lk_ulong)((k<<n_bits)|j);
					val[l]=1.0;
				}
			}
		} else {
			for(j=0;j<n_all;j++)	{
				a=a_set[i][j];
				if(!a) continue;
				if((cm[X_MAT])&(1<<j)) f=lf[X_MAT];
				else f=fq[j];
				for(k=0;k<n_all;k++) if(a&(1<<k)) {
					if((cm[X_PAT])&(1<<k)) f1=lf[X_PAT];
					else f1=fq[k];
					l=(lk_ulong)((k<<n_bits)|j);
					val[l]=f*f1;
				}
			}
		}
	}
	/* Add penetrances */
	if(pen) pen(val,i,locus,n_all,n_bits);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_simple_sample"
/* Sample a nuclear family */
static double loki_simple_sample(const struct Simple_Element *element,const int locus,pen_func pen)
{
	int ids,idd,i,j,k,l,m,n,fsp=0,n_off,*off,kid,*lump;
	double *tp,p1,p2,z,prob=0.0,pp[4];
	unsigned long a,b,cm[2],*cm1[2];
		  
	ids=element->sire-1;
	idd=element->dam-1;
	off=element->off;
	n_off=element->n_off;
	if(peel_trace) {
		printf("%s(): Trace - ",FUNC_NAME);
		if(family_id) {
			print_orig_family(stdout,off[0]+1,0);
			fputc(' ',stdout);
		}
		print_orig_id1(stdout,ids+1);
		fputc(',',stdout);
		print_orig_id1(stdout,idd+1);
		fputc(' ',stdout);
		for(i=0;i<n_off;i++) {
			fputc(i?',':'(',stdout);
			print_orig_id1(stdout,off[i]);
		}
		fputs(")\n",stdout);
	}
	if(id_array[ids].flag&SAMPLED_MAT) {
		if(!(id_array[ids].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 paternal allele sampled\n");
		i=1;
		for(j=0;j<n_idx;j++) pval[j]=0.0;
		j=id_array[ids].allele[X_MAT]-1;
		k=id_array[ids].allele[X_PAT]-1;
		l=(k<<n_bits)|j;
		pval[l]=1.0;
		if(peel_trace&4) {
			for(z=0.0,j=0;j<n_idx;j++) z+=pval[l];
			printf("Sire total (a): %g\n",z);
		}
	} else {
		i=id_array[ids].ngens;
		get_par_probs(pval,ids,locus,pen);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=pval[k];
		for(k=0;k<n_idx;k++) pval[k]/=p1;
		prob+=log(p1);
		if(peel_trace&4) {
			printf("Sire total (b): %g\n",p1);
		}
	}
	if(id_array[idd].flag&SAMPLED_MAT) {
		if(!(id_array[idd].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 maternal allele sampled\n");
		for(j=0;j<n_idx;j++) mval[j]=0.0;
		j=id_array[idd].allele[X_MAT]-1;
		k=id_array[idd].allele[X_PAT]-1;
		l=(k<<n_bits)|j;
		mval[l]=1.0;
		if(peel_trace&4) {
			for(z=0.0,j=0;j<n_idx;j++) z+=pval[l];
			printf("Dam total (a): %g\n",z);
		}
	} else {
		i*=id_array[idd].ngens;
		get_par_probs(mval,idd,locus,pen);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=mval[k];
		for(k=0;k<n_idx;k++) mval[k]/=p1;
		prob+=log(p1);
		if(peel_trace&4) {
			printf("Dam total (b): %g\n",p1);
		}
	}
	cm1[0]=peel_pt1;
	cm1[1]=cm1[0]+n_off;
	for(m=0;m<n_off;m++) {
		kid=off[m]-1;
		cm[0]=req_set[0][locus][kid];
		cm[1]=req_set[1][locus][kid];
		if(id_array[kid].flag&SAMPLED_MAT) {
			cm1[X_MAT][m]=1<<(id_array[kid].allele[X_MAT]-1);
			cm1[X_PAT][m]=1<<(id_array[kid].allele[X_PAT]-1);
		} else {
			cm1[X_PAT][m]=id_array[kid].temp[X_PAT];
			cm1[X_MAT][m]=id_array[kid].temp[X_MAT];
		}
		if(cm[X_MAT]&cm1[X_MAT][m]) cm1[X_MAT][m]|=cm[X_MAT];
		if(cm[X_PAT]&cm1[X_PAT][m]) cm1[X_PAT][m]|=cm[X_PAT];
	}
	/* Construct set of possible parental genotype combinations */
	for(i=0;i<n_all;i++) {
		a=a_set[ids][i];
		if(!a || ((id_array[ids].flag&SAMPLED_MAT) && (id_array[ids].allele[X_MAT]!=i+1))) continue;
		for(j=0;j<n_all;j++) if(a&(1<<j)) {
			if((id_array[ids].flag&SAMPLED_PAT) && (id_array[ids].allele[X_PAT]!=j+1)) continue;
			cm[X_PAT]=(1<<i)|(1<<j);
			for(m=0;m<n_off;m++) if(!(cm1[X_PAT][m]&cm[X_PAT])) break;
			if(m<n_off) continue;
			for(k=0;k<n_all;k++) {
				b=a_set[idd][k];
				if(!b || ((id_array[idd].flag&SAMPLED_MAT) && (id_array[idd].allele[X_MAT]!=k+1))) continue;
				for(l=0;l<n_all;l++) if(b&(1<<l)) {
					if((id_array[idd].flag&SAMPLED_PAT) && (id_array[idd].allele[X_PAT]!=l+1)) continue;
					cm[X_MAT]=(1<<k)|(1<<l);
					for(m=0;m<n_off;m++) if(!(cm1[X_MAT][m]&cm[X_MAT])) break;
					if(m==n_off) {
						peel_fs[fsp].pat_gene[X_MAT]=i;
						peel_fs[fsp].pat_gene[X_PAT]=j;
						peel_fs[fsp].mat_gene[X_MAT]=k;
						peel_fs[fsp].mat_gene[X_PAT]=l;
						peel_fs[fsp++].p=pval[(j<<n_bits)|i]*mval[(l<<n_bits)|k];
					}
				}
			}
		}
	}
	if(!fsp)	{
		ABT_FUNC(" - Internal error - no parental combinations\n");
	}
	for(m=0;m<n_off;m++) {
		kid=off[m]-1;
		for(k=0;k<n_idx;k++) qval[k]=0.0;
		if(id_array[kid].flag&SAMPLED_MAT) {
			if(!(id_array[kid].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 parental allele sampled\n");
			j=id_array[kid].allele[X_MAT]-1;
			k=id_array[kid].allele[X_PAT]-1;
			l=(k<<n_bits)|j;
			qval[l]=1.0;
		} else {
			if((k=id_array[kid].rfp)>=0) {/* Insert Previously computed R_Func */
				for(j=0;j<rf[k].n_terms;j++) qval[rf[k].index[j]]=rf[k].p[j];
			} else {
				for(j=0;j<n_all;j++) {
					a=a_set[kid][j];
					if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
						i=(k<<n_bits)|j;
						qval[i]=1.0;
					}
				}
			}
		}
		if(pen) pen(qval,kid,locus,n_all,n_bits);
		tp=id_array[kid].tp;
		cm[0]=req_set[0][locus][kid];
		cm[1]=req_set[1][locus][kid];
		if(!(cm[0] || cm[1])) for(n=0;n<fsp;n++) {
			i=peel_fs[n].pat_gene[X_MAT];
			j=peel_fs[n].pat_gene[X_PAT];
			k=peel_fs[n].mat_gene[X_MAT];
			l=peel_fs[n].mat_gene[X_PAT];
			peel_fs[n].p*=
			  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
			  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
			  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
			  tp[X_MP_PP]*qval[(j<<n_bits)|l];
		} else {
			lump=id_array[kid].lumped;
			for(n=0;n<fsp;n++) {
				i=peel_fs[n].pat_gene[X_MAT];
				j=peel_fs[n].pat_gene[X_PAT];
				k=peel_fs[n].mat_gene[X_MAT];
				l=peel_fs[n].mat_gene[X_PAT];
				if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
				if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
				if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
				if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
				peel_fs[n].p*=
				  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
				  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
				  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
				  tp[X_MP_PP]*qval[(j<<n_bits)|l];
			}
		}
	}
	p1=0.0;
	for(n=0;n<fsp;n++) p1+=peel_fs[n].p;
	prob+=log(p1);
	do {
		z=ranf()*p1;
		p2=0.0;
		for(n=0;n<fsp;n++) {
			if(peel_fs[n].p>0.0)	{
				p2+=peel_fs[n].p;
				if(p2>=z) {
					id_array[ids].allele[X_MAT]=peel_fs[n].pat_gene[X_MAT]+1;
					id_array[ids].allele[X_PAT]=peel_fs[n].pat_gene[X_PAT]+1;
					id_array[idd].allele[X_MAT]=peel_fs[n].mat_gene[X_MAT]+1;
					id_array[idd].allele[X_PAT]=peel_fs[n].mat_gene[X_PAT]+1;
					id_array[ids].flag|=(SAMPLED_MAT|SAMPLED_PAT);
					id_array[idd].flag|=(SAMPLED_MAT|SAMPLED_PAT);
					break;
				}
			}
		}
	} while(n==fsp);
	for(m=0;m<n_off;m++) {
		kid=off[m]-1;
		if(id_array[kid].flag&SAMPLED_MAT) continue;
		for(k=0;k<n_idx;k++) qval[k]=0.0;
		if((k=id_array[kid].rfp)>=0) {/* Insert Previously computed R_Func */
			for(j=0;j<rf[k].n_terms;j++) qval[rf[k].index[j]]=rf[k].p[j];
		} else {
			for(j=0;j<n_all;j++) {
				a=a_set[kid][j];
				if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
					i=(k<<n_bits)|j;
					qval[i]=1.0;
				}
			}
		}
		if(pen) pen(qval,kid,locus,n_all,n_bits);
		/* transmission probs */
		tp=id_array[kid].tp;
		cm[0]=req_set[0][locus][kid];
		cm[1]=req_set[1][locus][kid];
		i=id_array[ids].allele[X_MAT]-1;
		j=id_array[ids].allele[X_PAT]-1;
		k=id_array[idd].allele[X_MAT]-1;
		l=id_array[idd].allele[X_PAT]-1;
		if((1<<i)&cm[X_PAT]) i=id_array[kid].lumped[X_PAT];
		if((1<<j)&cm[X_PAT]) j=id_array[kid].lumped[X_PAT];
		if((1<<k)&cm[X_MAT]) k=id_array[kid].lumped[X_MAT];
		if((1<<l)&cm[X_MAT]) l=id_array[kid].lumped[X_MAT];
		p1=(pp[X_MM_PM]=tp[X_MM_PM]*qval[(i<<n_bits)|k]);
		p1+=(pp[X_MM_PP]=tp[X_MM_PP]*qval[(j<<n_bits)|k]);
		p1+=(pp[X_MP_PM]=tp[X_MP_PM]*qval[(i<<n_bits)|l]);
		p1+=(pp[X_MP_PP]=tp[X_MP_PP]*qval[(j<<n_bits)|l]);
		if(p1<=0.0) ABT_FUNC("Internal error - zero probability for child sample\n");
		do {
			z=ranf()*p1;
			p2=0.0;
			for(n=0;n<4;n++) {
				if(pp[n]>0.0) {
					p2+=pp[n];
					if(z<=p2) break;
				}
			}
		} while(n==4);
		switch(n) {
		 case X_MM_PM:
			id_array[kid].allele[X_MAT]=k+1;
			id_array[kid].allele[X_PAT]=i+1;
			break;
		 case X_MM_PP:
			id_array[kid].allele[X_MAT]=k+1;
			id_array[kid].allele[X_PAT]=j+1;
			break;
		 case X_MP_PM:
			id_array[kid].allele[X_MAT]=l+1;
			id_array[kid].allele[X_PAT]=i+1;
			break;
		 case X_MP_PP:
			id_array[kid].allele[X_MAT]=l+1;
			id_array[kid].allele[X_PAT]=j+1;
			break;
		 default:
			ABT_FUNC("Internal error - illegal sample\n");
		}
		id_array[kid].flag|=(SAMPLED_MAT|SAMPLED_PAT);
	}
	k=element->out_index;
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "peel_to_par"
/* This is when all (non-pruned) children in a family have completely determined genotypes
 * so we can peel to the 2 parents separately */
static double peel_to_par(const struct Simple_Element *element,const int locus,pen_func pen)
{
	int ids,idd,i,j,k,l,m,n,fsp=0,fsp1=0,n_off,*off,kid,*all;
	double prob=0.0,*tpp,p1,pp[2],qp;
	unsigned long a,cm[2];

	ids=element->sire-1;
	idd=element->dam-1;
	if(id_array[ids].ngens>id_array[idd].ngens) i=id_array[ids].ngens;
	else i=id_array[idd].ngens;
	off=element->off;
	n_off=element->n_off;
	/* Construct set of possible parental genotypes */
	for(j=0;j<n_idx;j++) mval[j]=pval[j]=0.0;
	if((k=id_array[ids].rfp)>=0) { /* Insert Previously computed R_Func */
		a=(1<<n_bits)-1;
		for(j=0;j<rf[k].n_terms;j++) {
			i=(int)rf[k].index[j];
			peel_fs[fsp].pat_gene[X_MAT]=i&a;
			peel_fs[fsp++].pat_gene[X_PAT]=i>>n_bits;
			pval[i]=rf[k].p[j];
		}
	} else for(i=0;i<n_all;i++) {
		a=a_set[ids][i];
		if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
			peel_fs[fsp].pat_gene[X_MAT]=i;
			peel_fs[fsp++].pat_gene[X_PAT]=j;
			pval[(j<<n_bits)|i]=1.0;
		}
	}
	if((k=id_array[idd].rfp)>=0) { /* Insert Previously computed R_Func */
		a=(1<<n_bits)-1;
		for(j=0;j<rf[k].n_terms;j++) {
			i=(int)rf[k].index[j];
			peel_fs[fsp1].mat_gene[X_MAT]=i&a;
			peel_fs[fsp1++].mat_gene[X_PAT]=i>>n_bits;
			mval[i]=rf[k].p[j];
		}
	} else for(i=0;i<n_all;i++) {
		a=a_set[idd][i];
		if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
			peel_fs[fsp1].mat_gene[X_MAT]=i;
			peel_fs[fsp1++].mat_gene[X_PAT]=j;
			mval[(j<<n_bits)|i]=1.0;
		}
	}
	for(m=0;m<n_off;m++)	{
		kid=off[m]-1;
		if(id_array[kid].rfp>=0)
		  ABT_FUNC("Internal error - no R-Functions expected for this operation\n");
		if(id_array[kid].ngens>1)
		  ABT_FUNC("Internal error - offspring genotypes should be determined for this operation\n");
		all=id_array[kid].allele;
		for(k=0;k<n_idx;k++) qval[k]=1.0;
		if(pen) pen(qval,kid,locus,n_all,n_bits);
		qp=qval[((all[X_PAT]-1)<<n_bits)|(all[X_MAT]-1)];
		cm[0]=req_set[0][locus][kid];
		cm[1]=req_set[1][locus][kid];
		tpp=id_array[kid].tpp[X_PAT];
		for(n=0;n<fsp;n++) {
			i=peel_fs[n].pat_gene[X_MAT];
			j=peel_fs[n].pat_gene[X_PAT];
			if((1<<i)&cm[X_PAT]) i=id_array[kid].lumped[X_PAT];
			if((1<<j)&cm[X_PAT]) j=id_array[kid].lumped[X_PAT];
			p1=0.0;
			if(all[X_PAT]==(i+1)) p1=tpp[X_MAT];
			if(all[X_PAT]==(j+1)) p1+=tpp[X_PAT];
			pval[(j<<n_bits)|i]*=p1*qp;
		}
		tpp=id_array[kid].tpp[X_MAT];
		for(n=0;n<fsp1;n++) {
			i=peel_fs[n].mat_gene[X_MAT];
			j=peel_fs[n].mat_gene[X_PAT];
			if((1<<i)&cm[X_MAT]) i=id_array[kid].lumped[X_MAT];
			if((1<<j)&cm[X_MAT]) j=id_array[kid].lumped[X_MAT];
			p1=0.0;
			if(all[X_MAT]==(i+1)) p1=tpp[X_MAT];
			if(all[X_MAT]==(j+1)) p1+=tpp[X_PAT];
			mval[(j<<n_bits)|i]*=p1*qp;
		}
	}
	pp[X_MAT]=pp[X_PAT]=0.0;
	for(i=0;i<n_idx;i++) {
		pp[X_MAT]+=mval[i];
		pp[X_PAT]+=pval[i];
	}
	prob+=log(pp[X_MAT]*pp[X_PAT]);
	k=element->out_index;
	rf[k].n_ind=2;
	rf[k].n_terms=fsp;
	get_rf_memory(rf+k,fsp,MRK_MBLOCK);
	get_rf_memory(rf+k+1,fsp1,MRK_MBLOCK);
	for(n=0;n<fsp;n++) {
		i=peel_fs[n].pat_gene[X_MAT];
		j=peel_fs[n].pat_gene[X_PAT];
		l=(j<<n_bits)|i;
		rf[k].index[n]=(lk_ulong)l;
		rf[k].p[n]=pval[l]/pp[X_PAT];
	}
	id_array[ids].rfp=k++;
	rf[k].n_ind=2;
	rf[k].n_terms=fsp1;
	for(n=0;n<fsp1;n++) {
		i=peel_fs[n].mat_gene[X_MAT];
		j=peel_fs[n].mat_gene[X_PAT];
		l=(j<<n_bits)|i;
		rf[k].index[n]=(lk_ulong)l;
		rf[k].p[n]=mval[l]/pp[X_MAT];
	}
	id_array[idd].rfp=k;
	return prob;
}

/* Similar to get_par_probs, but for a trait locus */
static void get_trait_par_probs(double *val,const int i,const int locus)
{
	int j,k,fflag=0,l1,l2;
	double f,*fq=0;
	
	j=id_array[i].sire;
	if(j && !(id_array[j-1].pruned_flag[n_markers])) fflag=1;
	else {
		k=id_array[i].group-1;
		if(k<0) ABT_FUNC("Internal error - null group\n");
		fq=freq[k];
	}
	if((k=id_array[i].rfp)>=0) {/* Insert Previously computed R_Func */
		if(fflag) for(j=0;j<n_idx;j++) val[j]=rf[k].p[j];
		else {
			for(j=l1=0;l1<n_all;l1++) {
				f=fq[l1];
				for(l2=0;l2<n_all;l2++) {
					val[j]=rf[k].p[j]*fq[l2]*f;
					j++;
				}
			}
		}
	} else {
		if(fflag) {
			for(j=0;j<n_idx;j++) val[j]=1.0;
		} else {
			for(j=l1=0;l1<n_all;l1++) {
				f=fq[l1];
				for(l2=0;l2<n_all;l2++)	{
					val[j]=fq[l2]*f;
					j++;
				}
			}
		}
	}
	penetrance(val,i,locus,n_all,0);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_complex_peelop"
/* Similar to loki_complex_peelop, but for a trait locus */
static double loki_trait_complex_peelop(const struct Complex_Element *element,const int locus,const int s_flag)
{
	int i,j,k,k1,k2,k3,n_out,n_peel,n_inv,*inv,n_rf,n_ind,n_other,ef,ef1,sampling=0;
	int *gt_store,*other_ptr,*other_list,*rf_ptr;
	int *off_index[2],*fnd_list,n_fnd=0,all,*pen_list,n_pen,*temp_p[10];
	double max_terms,*tp,prob=0.0,z,p1,Konst=0.0,*p_rf;
	
	n_terms=0; /* No. non-zero terms in output R-Function */
	/* Get details about peeling operation */
	n_inv=element->n_involved; /* No. alleles involved in op */
	n_peel=element->n_peel; /* No. allles to peel out (absorb) */
	n_out=n_inv-n_peel; /* No. to appear in output R-Function */
	n_rf=element->n_rfuncs; /* No. input R-Functions */
	inv=element->involved; /* List of involved alleles (alleles to peel come first) */
	/* if s_flag is non-zero then we are doing a sampling run */
	/* if s_flag&OP_SAMPLING then we are on the reverse (sampling) phase */
	/* In any case, if !n_out and s_flag then we can sample */
	if(s_flag) {
		if(s_flag&OP_SAMPLING) sampling=1;
		else if(!n_out) sampling=1;
	}
	/* In a sampling operation we sample the peeled alleles conditional on
	 * R-Functions and on the pivot (output) alleles which have all already been sampled */
	if(peel_trace) {
		for(i=0;i<n_peel;i++) {
			if(i) fputc(',',stdout);
			print_orig_allele_id(stdout,inv[i]);
		}
		for(i=0;i<n_out;i++)	{
			fputc(i?',':'|',stdout);
			print_orig_allele_id(stdout,inv[i+n_peel]);
		}
		if(s_flag) fputs(sampling?" S":" s",stdout);
		else fputs(" p",stdout);
	}
	if(sampling) {
		for(i=0;i<n_out;i++)	{
			j=inv[i+n_peel];
			if(j<0) {
				if(!(id_array[-j-1].flag&SAMPLED_PAT)) break;
			} else if(!(id_array[j-1].flag&SAMPLED_MAT)) break;
		}	
		if(i<n_out) ABT_FUNC("Internal error: Unsampled pivots\n");
	}
	if(peel_trace) fputc('\n',stdout);
	if(sampling) max_terms=log((double)n_all)*(double)n_peel;
	else max_terms=log((double)n_all)*(double)n_out;
	i=(int)(.5+exp(max_terms));
	if(!complex_out_p) {
		out_p_size=i+n_rf;
		if(!(complex_out_p=malloc(sizeof(double)*out_p_size))) ABT_FUNC(MMsg);
	}
	if(i+n_rf>out_p_size) {
		out_p_size=i+n_rf;
		if(!(complex_out_p=realloc(complex_out_p,sizeof(double)*out_p_size))) ABT_FUNC(MMsg);
	}
	p_rf=complex_out_p+i;
	for(j=0;j<i;j++) complex_out_p[j]=0.0;
	setup_complex_peel(element,sampling,temp_p,&n_other);
	gt_store=temp_p[0];
	other_list=temp_p[1];
	other_ptr=temp_p[2];
	off_index[0]=temp_p[5];
	off_index[1]=temp_p[6];
	pen_list=temp_p[7];
	fnd_list=temp_p[8];
	rf_ptr=temp_p[9];
	/* Compute masks - used for finding mutually 
	 * consistent terms from multiple input R-Functions */
	if(n_rf)	{
		for(i=0;i<n_rf;i++) {
			j=element->index[i];
			n_ind=rf[j].n_ind;
			k2=0;
			k3=1;
			for(k=n_ind-1;k>=0;k--)	{
				k1=rf[j].id_list[k];
				if(!gt_store[k1])	{
					k2|=1<<k1;
					k3*=n_all;
					gt_store[k1]=1;
				}
			}
			rf[j].mask1[sampling]=k2;
			rf[j].n_terms=k3;
		}
	}
	/* Make a list of alleles we need penetrances and/or founder probs for */
	n_pen=0;
	if(n_peel) {
		for(k=0;k<n_peel;k++) {
			j=abs(inv[k])-1;
			k1=id_array[j].sire;
			if(k1 && !id_array[k1-1].pruned_flag[n_markers]) continue;
			n_fnd++;
		}
		if(n_fnd) {
			n_fnd=0;
			for(k=0;k<n_peel;k++) {
				j=abs(inv[k])-1;
				k1=id_array[j].sire;
				if(k1 && !id_array[k1-1].pruned_flag[n_markers]) continue;
				fnd_list[n_fnd++]=k;
			}
		}
		for(k=0;k<n_peel;k++) {
			for(k1=k+1;k1<n_inv;k1++) if(inv[k1]== -inv[k]) break;
			if(k1==n_inv) continue;
			k1=abs(inv[k])-1;
			if(id_array[k1].res) pen_list[n_pen++]=k1;
		}
		if(n_pen) {
			if(!complex_pen) {
				if(!(complex_pen=malloc(sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
				max_pen1=n_pen*n_idx;
				if(!(complex_pen[0]=malloc(sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
				max_pen=n_pen;
			} else {
				if(n_pen>max_pen)	{
					max_pen=n_pen;
					if(!(complex_pen=realloc(complex_pen,sizeof(void *)*n_pen))) ABT_FUNC(MMsg);
				}
				if(n_pen*n_idx>max_pen1) {
					max_pen1=n_pen*n_idx;
					if(!(complex_pen[0]=realloc(complex_pen[0],sizeof(double)*max_pen1))) ABT_FUNC(MMsg);
				}
			}
			for(k1=0;k1<n_pen*n_idx;k1++) complex_pen[0][k1]=1.0;
			for(k1=1;k1<n_pen;k1++) complex_pen[k1]=complex_pen[k1-1]+n_idx;
			for(k=0;k<n_pen;k++) {
				penetrance(complex_pen[k],pen_list[k],locus,n_all,0);
				z=0.0;
				for(k1=0;k1<n_idx;k1++) z+=complex_pen[k][k1];
				if(z<=0.0) {
					if(!(s_flag&1)) return -DBL_MAX;
					ABT_FUNC("Zero probability!\n");
				}
				for(k1=0;k1<n_idx;k1++) complex_pen[k][k1]/=z;
				Konst+=log(z);
			}
		}
	}
	for(i=0;i<n_inv;i++) gt_store[i]=0;
	ef1=0;
	i=0;
	/* Main loop. */
	for(;;) {
		/* Fix allele types of sampled alleles */
		if(sampling) for(j=0;j<n_out;j++) {
			k=inv[j+n_peel];
			if(k>0) gt_store[j+n_peel]=id_array[k-1].allele[X_MAT];
			else gt_store[j+n_peel]=id_array[-k-1].allele[X_PAT];
		}
		/* Cycle through R-Functions */
		for(;i<n_rf;i++) {
			j=element->index[i];
			ef=0;
			n_ind=rf[j].n_ind;
			/* Find term consistent with what's already been set up */
			k3=rf_ptr[i];
			k2=0;
			for(k=n_ind-1;k>=0;k--)	{
				k1=rf[j].id_list[k];
				if(!gt_store[k1])	{
					gt_store[k1]=1+(k3%n_all);
					k3/=n_all;
				}
				k2=k2*n_all+gt_store[k1]-1;
			}
			p_rf[i]=rf[j].p[k2];
			if(p_rf[i]<=0.0) {/* Inconsistent - find the next 1 */
				for(;;) {
					k2=rf[j].mask1[sampling];
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(++rf_ptr[i]<rf[j].n_terms) break;
					if(!i) {
						ef1=1;
						break;
					}
					rf_ptr[i--]=0;
					j=element->index[i];
				}
				i--;
			}
			if(ef1) break;
		}
		if(ef1) break;
		for(ef=i=0;i<=n_other;i++) {
			if(i==n_other) {
				if(!ef) {
					/* Add Parent-off transmission probs. */
					p1=1.0;
					if(!ef) for(k=0;k<n_inv;k++) {
						if(element->flags[k]&HAD_P) { /* Sire-offspring */
							j= -inv[k]-1;
							tp=id_array[j].tpp[X_PAT];
							k3=gt_store[k];
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(k3==all) z=tp[X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(k3==all) z+=tp[X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
						} else if(element->flags[k]&HAD_M) { /* Dam-offspring */
							j=inv[k]-1;
							tp=id_array[j].tpp[X_MAT];
							k3=gt_store[k];
							z=0.0;
							all=gt_store[off_index[X_MAT][k]];
							if(k3==all) z=tp[X_MAT];
							all=gt_store[off_index[X_PAT][k]];
							if(k3==all) z+=tp[X_PAT];
							p1*=z;
							if(p1==0.0)	{
								ef=1;
								break;
							}
						}
					}
					/* Add penetrances and founder probs and store result */
					if(!ef) {
						for(k=0;k<n_rf;k++) p1*=p_rf[k];
						for(k1=0;k1<n_fnd;k1++) { /* Add founder frequencies */
							k=fnd_list[k1];
							j=abs(inv[k])-1;
							k2=id_array[j].group-1;
							if(k2<0) ABT_FUNC("OOOK!\n");
							p1*=freq[k2][gt_store[k]-1];
						}
						/* Add penetrances for those being peeled if required */
 						if(n_pen) {
							for(k=0;k<n_inv;k++)	{
								k1=inv[k];
								if(k1<0) id_array[-k1-1].allele[X_PAT]=gt_store[k];
								else id_array[k1-1].allele[X_MAT]=gt_store[k];
							}
							for(k=0;k<n_pen;k++)	{
								k1=pen_list[k];
								k2=((id_array[k1].allele[X_PAT]-1)<<n_bits)|(id_array[k1].allele[X_MAT]-1);
								p1*=complex_pen[k][k2];
							}
						}
						prob+=p1;
						k1=0;
						if(sampling) for(k=n_peel-1;k>=0;k--) k1=k1*n_all+gt_store[k]-1;
						else for(k=n_inv-1;k>=n_peel;k--) k1=k1*n_all+gt_store[k]-1;
						complex_out_p[k1]+=p1;
						k1=0;
						for(k=n_inv-1;k>=0;k--) k1=k1*n_all+gt_store[k]-1;
						ef=1;
					}
				}
			}
			/* Find next combination */
			if(ef) {
				i--;
				while(i>=0)	{
					other_ptr[i]++;
					gt_store[other_list[i]]=0;
					if(other_ptr[i]<n_all) break;
					other_ptr[i--]=0;
				}
				i--;
				if(i<-1) break;
				ef=0;
			} else gt_store[other_list[i]]=other_ptr[i]+1;
		}
		for(i=0;i<n_other;i++) gt_store[other_list[i]]=0;
		/* Cycle through R-Functions */
		if(n_rf)	{
			i=n_rf-1;
			do	{
				rf_ptr[i]++;
				j=element->index[i];
				k2=rf[j].mask1[sampling];
				if(k2) {
					k=0;
					while(k2) {
						if(k2&1) gt_store[k]=0;
						k++;
						k2>>=1;
					}
					if(rf_ptr[i]<rf[j].n_terms) break;
				}
				rf_ptr[i--]=0;
			} while(i>=0);
			if(i>=0) ef=0;
		} else i=0;
		if(ef) break;
	}
	/* All valid combinations have been visited.  Have we found any ? */
	if(prob<=0.0) {
		if(!(s_flag&1)) return -DBL_MAX;
		ABT_FUNC("Zero probability!\n");
	}
	n_terms=(int)(.5+exp(max_terms));
	/* If sampling then sample from output function */
	if(sampling) {
		do {
			z=ranf()*prob;
			p1=0.0;
			for(k=0;k<n_terms;k++) {
				if(complex_out_p[k]>0.0) {
					p1+=complex_out_p[k];
					if(z<=p1) break;
				}
			}
		} while(k==n_terms);
		for(k1=0;k1<n_peel;k1++) {
			gt_store[k1]=(k%n_all)+1;
			k/=n_all;
		}
		for(i=0;i<n_peel;i++) {
			j=inv[i];
			if(j>0) {
				id_array[j-1].allele[X_MAT]=gt_store[i];
				id_array[j-1].flag|=SAMPLED_MAT;
			} else {
				id_array[-j-1].allele[X_PAT]=gt_store[i];
				id_array[-j-1].flag|=SAMPLED_PAT;
			}
		}
	} else { /* Otherwise normalize and store */
		i=element->out_index;
		if(i>=0)	{
			get_rf_memory(rf+i,n_terms,TRT_MBLOCK);
			for(j=0;j<n_terms;j++) rf[i].p[j]=complex_out_p[j]/prob;
		}
	}
	return log(prob)+Konst;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_simple_sample"
/* Similar to loki_simple_sample, but for a trait locus */
static double loki_trait_simple_sample(const struct Simple_Element *element,const int locus,const int s_flag)
{
	int ids,idd,i,j,k,l,m,n,n_off,*off,kid;
	double *tp,p1,p2,z,prob=0.0,pp[4];
		  
	ids=element->sire-1;
	idd=element->dam-1;
	off=element->off;
	n_off=element->n_off;
	if(peel_trace) {
		printf("%s(): Trace - ",FUNC_NAME);
		if(family_id) {
			print_orig_family(stdout,off[0]+1,0);
			fputc(' ',stdout);
		}
		print_orig_id1(stdout,ids+1);
		fputc(',',stdout);
		print_orig_id1(stdout,idd+1);
		fputc(' ',stdout);
		for(i=0;i<n_off;i++) {
			fputc(i?',':'(',stdout);
			print_orig_id1(stdout,off[i]);
		}
		fputs(")\n",stdout);
	}
	if(id_array[ids].flag&SAMPLED_MAT) {
		if(!(id_array[ids].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 paternal allele sampled\n");
		for(j=0;j<n_idx;j++) pval[j]=0.0;
		j=id_array[ids].allele[X_MAT]-1;
		k=id_array[ids].allele[X_PAT]-1;
		pval[k*n_all+j]=1.0;
	} else {
		get_trait_par_probs(pval,ids,locus);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=pval[k];
		if(p1<=0.0)	{
			if(!(s_flag&(1|OP_SAMPLING))) return -DBL_MAX;
			ABT_FUNC("Zero probability in sampling operation\n");
		}
		for(k=0;k<n_idx;k++) pval[k]/=p1;
		prob+=log(p1);
	}
	if(id_array[idd].flag&SAMPLED_MAT) {
		if((!id_array[idd].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 maternal allele sampled\n");
		for(j=0;j<n_idx;j++) mval[j]=0.0;
		j=id_array[idd].allele[X_MAT]-1;
		k=id_array[idd].allele[X_PAT]-1;
		mval[k*n_all+j]=1.0;
	} else {
		get_trait_par_probs(mval,idd,locus);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=mval[k];
		if(p1<=0.0)	{
			if(!(s_flag&(1|OP_SAMPLING))) return -DBL_MAX;
			ABT_FUNC("Zero probability in sampling operation\n");
		}
		for(k=0;k<n_idx;k++) mval[k]/=p1;
		prob+=log(p1);
	}
	i=n_idx*n_idx;
	for(j=0;j<i;j++) peel_famval[j]=0.0;
	for(n=0;n<n_idx;n++) {
		p1=pval[n];
		if(p1<=0.0) continue;
		for(m=0;m<n_idx;m++) peel_famval[n*n_idx+m]=p1*mval[m];
	}
	for(m=0;m<n_off;m++) {
		kid=off[m]-1;
		if(id_array[kid].flag&SAMPLED_MAT) {
			for(k=0;k<n_idx;k++) qval[k]=0.0;
			if(!(id_array[kid].flag&SAMPLED_PAT)) ABT_FUNC("Internal error: 1 parental allele sampled\n");
			j=id_array[kid].allele[X_MAT]-1;
			k=id_array[kid].allele[X_PAT]-1;
			qval[k*n_all+j]=1.0;
		} else {
			if((k=id_array[kid].rfp)>=0) { /* Insert Previously computed R_Func */
				for(j=0;j<n_idx;j++) qval[j]=rf[k].p[j];
			} else for(j=0;j<n_idx;j++) qval[j]=1.0;
			penetrance(qval,kid,locus,n_all,0);
			for(p1=0.0,j=0;j<n_idx;j++) p1+=qval[j];
			if(p1<=0.0)	{
				if(!(s_flag&(1|OP_SAMPLING))) return -DBL_MAX;
				ABT_FUNC("Zero probability in sampling operation\n");
			}
			for(j=0;j<n_idx;j++) qval[j]/=p1;
			prob+=log(p1);
		}
		tp=id_array[kid].tp;
		for(n=0,j=0;j<n_all;j++) {
			for(i=0;i<n_all;i++)	{
				for(l=0;l<n_all;l++)	{
					for(k=0;k<n_all;k++)	{
						peel_famval[n++]*=
						  tp[X_MM_PM]*qval[i*n_all+k]+
						  tp[X_MM_PP]*qval[j*n_all+k]+
						  tp[X_MP_PM]*qval[i*n_all+l]+
						  tp[X_MP_PP]*qval[j*n_all+l];
					}
				}
			}
		}
	}
	p1=0.0;
	m=n_idx*n_idx;
	for(n=0;n<m;n++) p1+=peel_famval[n];
	if(p1<=0.0) {
		if(!(s_flag&(1|OP_SAMPLING))) return -DBL_MAX;
		ABT_FUNC("Zero probability in sampling operation\n");
	}
	prob+=log(p1);
	do {
		z=ranf()*p1;
		p2=0.0;
		for(m=n=0,j=0;j<n_all;j++) {
			for(i=0;i<n_all;i++) {
				for(l=0;l<n_all;l++) {
					for(k=0;k<n_all;k++) {
						p1=peel_famval[n++];
						if(p1<=0.0) continue;
						p2+=p1;
						if(p2>=z) {
							m=1;
							id_array[ids].allele[X_MAT]=i+1;
							id_array[ids].allele[X_PAT]=j+1;
							id_array[idd].allele[X_MAT]=k+1;
							id_array[idd].allele[X_PAT]=l+1;
							id_array[ids].flag|=(SAMPLED_MAT|SAMPLED_PAT);
							id_array[idd].flag|=(SAMPLED_MAT|SAMPLED_PAT);
							i=j=k=l=n_all;
							break;
						}
					}
				}
			}
		}
	} while(!m);
	for(m=0;m<n_off;m++)	{
		kid=off[m]-1;
		if(id_array[kid].flag&SAMPLED_MAT) continue;
		if((k=id_array[kid].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) qval[j]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) qval[j]=1.0;
		penetrance(qval,kid,locus,n_all,0);
		for(p1=0.0,j=0;j<n_idx;j++) p1+=qval[j];
		if(p1<=0.0) ABT_FUNC("Internal error - no offspring combination possible\n");
		for(j=0;j<n_idx;j++) qval[j]/=p1;
		/* transmission probs */
		tp=id_array[kid].tp;
		i=id_array[ids].allele[X_MAT]-1;
		j=id_array[ids].allele[X_PAT]-1;
		k=id_array[idd].allele[X_MAT]-1;
		l=id_array[idd].allele[X_PAT]-1;
		p1=(pp[X_MM_PM]=tp[X_MM_PM]*qval[i*n_all+k]);
		p1+=(pp[X_MM_PP]=tp[X_MM_PP]*qval[j*n_all+k]);
		p1+=(pp[X_MP_PM]=tp[X_MP_PM]*qval[i*n_all+l]);
		p1+=(pp[X_MP_PP]=tp[X_MP_PP]*qval[j*n_all+l]);
		if(p1<=0.0) ABT_FUNC("Internal error - no offspring combination possible\n");
		do {
			z=ranf()*p1;
			p2=0.0;
			for(n=0;n<4;n++) {
				if(pp[n]>0.0) {
					p2+=pp[n];
					if(z<=p2) break;
				}
			}
		} while(n==4);
		switch(n) {
		 case X_MM_PM:
			id_array[kid].allele[X_MAT]=k+1;
			id_array[kid].allele[X_PAT]=i+1;
			break;
		 case X_MM_PP:
			id_array[kid].allele[X_MAT]=k+1;
			id_array[kid].allele[X_PAT]=j+1;
			break;
		 case X_MP_PM:
			id_array[kid].allele[X_MAT]=l+1;
			id_array[kid].allele[X_PAT]=i+1;
			break;
		 case X_MP_PP:
			id_array[kid].allele[X_MAT]=l+1;
			id_array[kid].allele[X_PAT]=j+1;
			break;
		 default:
			ABT_FUNC("Internal error - illegal sample\n");
		}
		id_array[kid].flag|=(SAMPLED_MAT|SAMPLED_PAT);
	}
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_trait_simple_peelop"
/* Similar to loki_simple_peelop, but for a trait locus */
static double loki_trait_simple_peelop(const struct Simple_Element *element,const int locus,const int s_flag)
{
	int ids,idd,i,j,k,l,m,n,n1,pivot,n_off,*off,kid,of=0;
	double prob=0.0,*tp,p1,z;
		  
	pivot=element->pivot-1;
	if(pivot== -3) ABT_FUNC("Internal error - this peeling operation should not occur with a trait locus\n");
	ids=element->sire-1;
	idd=element->dam-1;
	off=element->off;
	n_off=element->n_off;
	if(peel_trace) {
		printf("%s(): Trace - ",FUNC_NAME);
		if(family_id) {
			print_orig_family(stdout,off[0]+1,0);
			fputc(' ',stdout);
		}
		print_orig_id1(stdout,ids+1);
		fputc(',',stdout);
		print_orig_id1(stdout,idd+1);
		fputc(' ',stdout);
		for(i=0;i<n_off;i++) {
			fputc(i?',':'(',stdout);
			print_orig_id1(stdout,off[i]);
		}
		fputs(") -> ",stdout);
		print_orig_id1(stdout,pivot+1);
		fputc('\n',stdout);
	}
	if(ids<0) { /* Peeling singletons */
		for(m=0;m<n_off;m++)	{
			kid=off[m]-1;
			get_trait_par_probs(pval,kid,locus);
			p1=0.0;
			for(k=0;k<n_idx;k++) p1+=pval[k];
			if(p1<=0.0) {
				if(!(s_flag&1)) return -DBL_MAX;
				ABT_FUNC("Zero probability in peeling operation\n");
			}
			prob+=log(p1);
			if(s_flag) {
				do {
					z=p1*ranf();
					p1=0.0;
					for(i=0;i<n_idx;i++) if(pval[i])	{
						p1+=pval[i];
						if(z<=p1) break;
					}
				} while(i==n_idx);
				id_array[kid].allele[X_MAT]=1+(i%n_all);
				id_array[kid].allele[X_PAT]=1+((i/n_all)%n_all);
				id_array[kid].flag|=(SAMPLED_MAT|SAMPLED_PAT);
			} else for(k=0;k<n_idx;k++) pval[k]/=p1;
		}
		return prob;
	}
	if(s_flag && !(element->pivot)) 
	  return loki_trait_simple_sample(element,locus,s_flag);
	i=n_idx*n_idx;
	if(idd!=pivot && pivot!= -2) {
		get_trait_par_probs(mval,idd,locus);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=mval[k];
		if(p1<=0.0)	{
			if(!(s_flag&1)) return -DBL_MAX;
			ABT_FUNC("Zero probability in peeling operation\n");
		}
		for(k=0;k<n_idx;k++) mval[k]/=p1;
		prob+=log(p1);
	} else {
		if((k=id_array[idd].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) mval[j]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) mval[j]=1.0;
	}
	if(ids!=pivot && pivot!= -2) {
		get_trait_par_probs(pval,ids,locus);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=pval[k];
		if(p1<=0.0)	{
			if(!(s_flag&1)) return -DBL_MAX;
			ABT_FUNC("Zero probability in peeling operation\n");
		}
		for(k=0;k<n_idx;k++) pval[k]/=p1;
		prob+=log(p1);
	} else {
		if((k=id_array[ids].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) pval[j]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) pval[j]=1.0;
	}
	for(n=i=0;i<n_idx;i++) {
		p1=pval[i];
		for(k=0;k<n_idx;k++) peel_famval[n++]=p1*mval[k];
	}
	for(m=0;m<n_off;m++)	{
		kid=off[m]-1;
		if(kid==pivot) continue;
		if((k=id_array[kid].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) qval[j]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) qval[j]=1.0;
		penetrance(qval,kid,locus,n_all,0);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=qval[k];
		if(p1<=0.0) {
			if(!(s_flag&1)) return -DBL_MAX;
			ABT_FUNC("Zero probability in peeling operation\n");
		}
		for(k=0;k<n_idx;k++) qval[k]/=p1;
		prob+=log(p1);
		tp=id_array[kid].tp;
		for(n=j=0;j<n_all;j++) {
			for(i=0;i<n_all;i++)	{
				for(l=0;l<n_all;l++)	{
					for(k=0;k<n_all;k++)	{
						peel_famval[n++]*=
						  tp[X_MM_PM]*qval[i*n_all+k]+
						  tp[X_MM_PP]*qval[j*n_all+k]+
						  tp[X_MP_PM]*qval[i*n_all+l]+
						  tp[X_MP_PP]*qval[j*n_all+l];
					}
				}
			}
		}
	}
	if(pivot== -2)	{
		p1=0.0;
		m=n_idx*n_idx;
		for(n=0;n<m;n++) {
			p1+=peel_famval[n];
			n1++;
		}
		if(p1<=0.0) {
			if(!(s_flag&1)) return -DBL_MAX;
			ABT_FUNC("Zero probability in peeling operation\n");
		}
		prob+=log(p1);
		k=element->out_index;
		rf[k].n_ind=4;
		get_rf_memory(rf+k,m,TRT_MBLOCK);
		for(n=0;n<m;n++) rf[k].p[n]=peel_famval[n]/p1;
		return prob;
	}
	if(pivot>=0 && pivot!=ids && pivot!=idd) {
		tp=id_array[pivot].tp;
		if((k=id_array[pivot].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) pval[j]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) pval[j]=1.0;
		of=1;
	} else tp=0;
	for(k=0;k<n_idx;k++) qval[k]=0.0;
	p1=0.0;
	for(n=j=0;j<n_all;j++) {
		for(i=0;i<n_all;i++) {
			for(l=0;l<n_all;l++)	{
				for(k=0;k<n_all;k++)	{
					if(idd==pivot) qval[l*n_all+k]+=peel_famval[n];
					else if(ids==pivot) qval[j*n_all+i]+=peel_famval[n];
					else if(pivot<0) p1+=peel_famval[n];
					else {
						p1=peel_famval[n];
						qval[i*n_all+k]+=tp[X_MM_PM]*p1;
						qval[j*n_all+k]+=tp[X_MM_PP]*p1;
						qval[i*n_all+l]+=tp[X_MP_PM]*p1;
						qval[j*n_all+l]+=tp[X_MP_PP]*p1;
					}
					n++;
				}
			}
		}
	}
	if(pivot>=0) {
		p1=0.0;
		if(of) for(j=0;j<n_idx;j++) p1+=(qval[j]*=pval[j]);
		else for(j=0;j<n_idx;j++) p1+=qval[j];
		if(p1<=0.0) {
			if(!(s_flag&1)) return -DBL_MAX;
			ABT_FUNC("Zero probability in peeling operation\n");
		}
		prob+=log(p1);
		k=element->out_index;
		id_array[pivot].rfp=k;
		rf[k].n_ind=2;
		get_rf_memory(rf+k,n_idx,TRT_MBLOCK);
		for(j=0;j<n_idx;j++) rf[k].p[j]=qval[j]/p1;
	} else prob+=log(p1);
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "loki_simple_peelop"
/* Performs simple (i.e., nuclear family based) peeling operation */
static double loki_simple_peelop(const struct Simple_Element *element,const int locus,const int s_flag,pen_func pen)
{
	int ids,idd,i,j,k,l,m,n,n1,pivot,fsp=0,n_off,*off,kid,gt[4],of=0,*lump;
	double prob=0.0,*tp,p1,z;
	unsigned long a,b,cm[2];
		  
	pivot=element->pivot-1;
	if(pivot== -3) return peel_to_par(element,locus,pen);
	ids=element->sire-1;
	idd=element->dam-1;
	off=element->off;
	n_off=element->n_off;
	if(peel_trace) {
		printf("%s(): Trace - ",FUNC_NAME);
		if(family_id) {
			print_orig_family(stdout,off[0]+1,0);
			fputc(' ',stdout);
		}
		print_orig_id1(stdout,ids+1);
		fputc(',',stdout);
		print_orig_id1(stdout,idd+1);
		fputc(' ',stdout);
		for(i=0;i<n_off;i++) {
			fputc(i?',':'(',stdout);
			print_orig_id1(stdout,off[i]);
		}
		fputs(") -> ",stdout);
		print_orig_id1(stdout,pivot+1);
		fputc('\n',stdout);
	}
	if(ids<0) { /* Peeling singletons */
		for(m=0;m<n_off;m++)	{
			kid=off[m]-1;
			get_par_probs(pval,kid,locus,pen);
			p1=0.0;
			for(k=0;k<n_idx;k++) p1+=pval[k];
			if(p1<=0.0)	{
				if(!(s_flag&1)) return -DBL_MAX;
				ABT_FUNC("Zero probability in peeling operation\n");
			}
			prob+=log(p1);
			if(s_flag) {
				do {
					z=p1*ranf();
					p1=0.0;
					for(i=0;i<n_idx;i++) if(pval[i]) {
						p1+=pval[i];
						if(z<=p1) break;
					}
				} while(i==n_idx);
				id_array[kid].allele[X_MAT]=1+(i&mask);
				id_array[kid].allele[X_PAT]=1+((i>>n_bits)&mask);
				id_array[kid].flag|=(SAMPLED_MAT|SAMPLED_PAT);
			} else for(k=0;k<n_idx;k++) pval[k]/=p1;
		}
		return prob;
	}
	if(s_flag && !(element->pivot)) return loki_simple_sample(element,locus,pen);
	i=id_array[ids].ngens*id_array[idd].ngens;
	if(idd!=pivot && pivot!= -2) {
		get_par_probs(mval,idd,locus,pen);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=mval[k];
		for(k=0;k<n_idx;k++) mval[k]/=p1;
		prob+=log(p1);
	} else {
		if((k=id_array[idd].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) mval[j]=0.0;
			for(j=0;j<rf[k].n_terms;j++) mval[rf[k].index[j]]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) mval[j]=1.0;
	}
	if(ids!=pivot && pivot!= -2) {
		get_par_probs(pval,ids,locus,pen);
		p1=0.0;
		for(k=0;k<n_idx;k++) p1+=pval[k];
		for(k=0;k<n_idx;k++) pval[k]/=p1;
		prob+=log(p1);
	} else {
		if((k=id_array[ids].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<n_idx;j++) pval[j]=0.0;
			for(j=0;j<rf[k].n_terms;j++) pval[rf[k].index[j]]=rf[k].p[j];
		} else for(j=0;j<n_idx;j++) pval[j]=1.0;
	}
	/* Construct set of possible parental genotype combinations */
	for(i=0;i<n_all;i++) {
		a=a_set[ids][i];
		if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
			p1=pval[(j<<n_bits)|i];
			for(k=0;k<n_all;k++)	{
				b=a_set[idd][k];
				if(b) for(l=0;l<n_all;l++) if(b&(1<<l)) {
					peel_fs[fsp].pat_gene[X_MAT]=i;
					peel_fs[fsp].pat_gene[X_PAT]=j;
					peel_fs[fsp].mat_gene[X_MAT]=k;
					peel_fs[fsp].mat_gene[X_PAT]=l;
					peel_fs[fsp++].p=p1*mval[(l<<n_bits)|k];
				}
			}
		}
	}
	/* Add contributions from non-pivot offspring */
	for(m=0;m<n_off;m++) {
		kid=off[m]-1;
		if(kid==pivot) continue;
		for(k=0;k<n_idx;k++) qval[k]=0.0;
		if((k=id_array[kid].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<rf[k].n_terms;j++) qval[rf[k].index[j]]=rf[k].p[j];
		} else {
			for(j=0;j<n_all;j++)	{
				a=a_set[kid][j];
				if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) qval[(k<<n_bits)|j]=1.0;
			}
		}
		if(pen) pen(qval,kid,locus,n_all,n_bits);
		tp=id_array[kid].tp;
		cm[0]=req_set[0][locus][kid];
		cm[1]=req_set[1][locus][kid];
		if(!(cm[0] || cm[1])) for(n=0;n<fsp;n++) {
			if(peel_fs[n].p<=0.0) continue;
			i=peel_fs[n].pat_gene[X_MAT];
			j=peel_fs[n].pat_gene[X_PAT];
			k=peel_fs[n].mat_gene[X_MAT];
			l=peel_fs[n].mat_gene[X_PAT];
			peel_fs[n].p*=
			  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
			  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
			  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
			  tp[X_MP_PP]*qval[(j<<n_bits)|l];
		} else {
			lump=id_array[kid].lumped;
			for(n=0;n<fsp;n++) {
				if(peel_fs[n].p<=0.0) continue;
				i=peel_fs[n].pat_gene[X_MAT];
				j=peel_fs[n].pat_gene[X_PAT];
				k=peel_fs[n].mat_gene[X_MAT];
				l=peel_fs[n].mat_gene[X_PAT];
				if((1<<i)&cm[X_PAT]) i=lump[X_PAT];
				if((1<<j)&cm[X_PAT]) j=lump[X_PAT];
				if((1<<k)&cm[X_MAT]) k=lump[X_MAT];
				if((1<<l)&cm[X_MAT]) l=lump[X_MAT];
				peel_fs[n].p*=
				  tp[X_MM_PM]*qval[(i<<n_bits)|k]+
				  tp[X_MM_PP]*qval[(j<<n_bits)|k]+
				  tp[X_MP_PM]*qval[(i<<n_bits)|l]+
				  tp[X_MP_PP]*qval[(j<<n_bits)|l];
			}
		}
	}
	if(pivot== -2) { /* Peeling to joint on both parents */
		p1=0.0;
		for(n1=n=0;n<fsp;n++) {
			if(peel_fs[n].p<=0.0) continue;
			p1+=peel_fs[n].p;
			n1++;
		}
		prob+=log(p1);
		k=element->out_index;
		rf[k].n_ind=4;
		rf[k].n_terms=n1;
		get_rf_memory(rf+k,n1,MRK_MBLOCK);
		for(n1=n=0;n<fsp;n++) {
			if(peel_fs[n].p<=0.0) continue;
			gt[0]=peel_fs[n].mat_gene[X_MAT]+1;
			gt[1]=peel_fs[n].mat_gene[X_PAT]+1;
			gt[2]=peel_fs[n].pat_gene[X_MAT]+1;
			gt[3]=peel_fs[n].pat_gene[X_PAT]+1;
			rf[k].index[n1]=get_index1(4,gt);
			rf[k].p[n1++]=peel_fs[n].p/p1;
		}
		return prob;
	}
	/* If pivot is an offspring, bring in previous R-Function and zero out
	 * illegal genotypes */
	if(pivot>=0 && pivot!=ids && pivot!=idd) {
		tp=id_array[pivot].tp;
		cm[0]=req_set[0][locus][pivot];
		cm[1]=req_set[1][locus][pivot];
		for(k=0;k<n_idx;k++) pval[k]=0.0;
		if((k=id_array[pivot].rfp)>=0) { /* Insert Previously computed R_Func */
			for(j=0;j<rf[k].n_terms;j++) pval[rf[k].index[j]]=rf[k].p[j];
		} else {
			for(j=0;j<n_all;j++) {
				a=a_set[pivot][j];
				if(a) for(k=0;k<n_all;k++) if(a&(1<<k)) {
					pval[(k<<n_bits)|j]=1.0;
				}
			}
		}
		of=1;
	} else tp=0;
	/* Assemble output function in qval */
	for(k=0;k<n_idx;k++) qval[k]=0.0;
	p1=0.0;
	for(n=0;n<fsp;n++) {
		if(peel_fs[n].p<=0.0) continue;
		i=peel_fs[n].pat_gene[X_MAT];
		j=peel_fs[n].pat_gene[X_PAT];
		k=peel_fs[n].mat_gene[X_MAT];
		l=peel_fs[n].mat_gene[X_PAT];
		if(idd==pivot) qval[(l<<n_bits)|k]+=peel_fs[n].p;
		else if(ids==pivot) qval[(j<<n_bits)|i]+=peel_fs[n].p;
		else if(pivot<0) p1+=peel_fs[n].p;
		else {
			p1=peel_fs[n].p;
			if((1<<i)&cm[X_PAT]) i=id_array[pivot].lumped[X_PAT];
			if((1<<j)&cm[X_PAT]) j=id_array[pivot].lumped[X_PAT];
			if((1<<k)&cm[X_MAT]) k=id_array[pivot].lumped[X_MAT];
			if((1<<l)&cm[X_MAT]) l=id_array[pivot].lumped[X_MAT];
			qval[(i<<n_bits)|k]+=tp[X_MM_PM]*p1;
			qval[(j<<n_bits)|k]+=tp[X_MM_PP]*p1;
			qval[(i<<n_bits)|l]+=tp[X_MP_PM]*p1;
			qval[(j<<n_bits)|l]+=tp[X_MP_PP]*p1;
		}
	}
	/* If we have a pivot then normalize and store output function */
	if(pivot>=0) {
		i=0;
		p1=0.0;
		if(of) {
			for(j=0;j<n_idx;j++) if((qval[j]*=pval[j])>0.0) {
				i++;
				p1+=qval[j];
			}
		} else for(j=0;j<n_idx;j++) if(qval[j]>0.0) {
			i++;
			p1+=qval[j];
		}
		prob+=log(p1);
		k=element->out_index;
		id_array[pivot].rfp=k;
		rf[k].n_ind=2;
		rf[k].n_terms=i;
		if(!i) ABT_FUNC("Internal error - zero possible combinations\n");
		get_rf_memory(rf+k,i,MRK_MBLOCK);
		for(i=j=0;j<n_idx;j++) if(qval[j]) {
			rf[k].p[i]=qval[j]/p1;
			rf[k].index[i++]=(lk_ulong)j;
		}
	} else prob+=log(p1);
	return prob;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "peel_alloc"
void peel_alloc(void)
{
	int i,j=0,k=2,k1=0;

	for(i=0;i<n_markers;i++) {
		j=marker[i].n_alleles;
		if(j>k) k=j;
	}
	if(!k) return;
	if(!(freq=malloc(sizeof(void *)*2*n_genetic_groups))) ABT_FUNC(MMsg);
	if(!(freq[0]=malloc(sizeof(double)*k*2*n_genetic_groups))) ABT_FUNC(MMsg);
	for(i=1;i<n_genetic_groups*2;i++) freq[i]=freq[i-1]+k;
   /* Note the 16 below should change if more than diallelic trait loci are fitted */
	if(!(peel_famval=malloc(sizeof(double)*16))) ABT_FUNC(MMsg);
	if(!(peel_alls=malloc(sizeof(int)*k))) ABT_FUNC(MMsg);
	k1=k*k;
	if(!(peel_fs=malloc(sizeof(struct fset)*k1*k1))) ABT_FUNC(MMsg);
	j=num_bits(k);
	j=1<<(j+j);
	if(!(qval=malloc(sizeof(double)*3*j))) ABT_FUNC(MMsg);
	pval=qval+j;
	mval=pval+j;
	if(comp_npeel) {
		for(k1=i=0;i<n_comp;i++) {
			j=comp_npeel[i];
			if(j>k1) k1=j;
		}
		if(k1) if(!(peel_list=malloc(sizeof(void *)*k1))) ABT_FUNC(MMsg);
	}
	if(max_peel_off) if(!(peel_pt1=malloc(sizeof(long)*max_peel_off*2))) ABT_FUNC(MMsg);
	first_mem_block[MRK_MBLOCK]=get_new_memblock(mb_size,MRK_MBLOCK);
	first_mem_block[TRT_MBLOCK]=get_new_memblock(mb_size,TRT_MBLOCK);
}

void peel_dealloc(void)
{
	int k;
	struct peel_mem_block *p,*p1;
	
	if(freq) {
		if(freq[0]) free(freq[0]);
		free(freq);
	}
	if(peel_famval) free(peel_famval);
	if(peel_alls) free(peel_alls);
	if(qval) free(qval);
	if(peel_list) free(peel_list);
	if(peel_fs) free(peel_fs);
	if(peel_pt1) free(peel_pt1);
	for(k=0;k<2;k++) {
		p1=first_mem_block[k];
		while(p1) {
			p=p1->next;
			if(p1->index) free(p1->index);
			if(p1->val) free(p1->val);
			free(p1);
			p1=p;
		}
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "peel_locus"
/* Peels locus perm[idx] where the n_loci entries in perm are all in 1 linkage group.
 * If sample_flag then the locus is to be sampled */
double peel_locus(const int *perm,int idx,int n_loci,int sample_flag)
{
	struct Peelseq_Head *pp;
	struct Simple_Element *simple_em;
   struct Complex_Element *complex_em; 
	int comp,i,i1,j,k,k1,k2,k3,nn_all,n_peel_ops,ids,idd,allele,n_rf,grp;
	int k1a,k2a,rec,nrec,mtype,locus,sample_freq=0,trn[2];
   double like=0.0,z,z1,theta,Mtp[2],Ptp[2],*pos,*eff,**count,*tpp;
	signed char **seg;
	lk_ulong a,b,b1,c,lump;
	pen_func *pen=0;
	
	locus=perm[idx];
	sample_flag&=~OP_SAMPLING;
	if(locus<0) { /* Trait locus */
		nn_all=tlocus[-1-locus].n_alleles;
		n_bits=num_bits(nn_all);
		pos=tlocus[-1-locus].pos;
		seg=tlocus[-1-locus].seg;
		if(tlocus[-1-locus].flag&TL_LINKED) {
			j=tlocus[-1-locus].link_group;
			linktype=linkage[j].type;
		} else linktype=LINK_AUTO;
	} else { /* Marker locus */
		nn_all=marker[locus].n_alleles;
		a_set=all_set[locus];
		pos=marker[locus].pos;
		sample_freq=0;
		for(k=0;k<n_genetic_groups;k++) {
			for(i=j=0;i<nn_all;i++) if(marker[locus].freq_set[k][i]!=1) j++;
			sample_freq|=(j>1)?1:0;
		}
		seg=marker[locus].seg;
		j=marker[locus].link_group;
		linktype=linkage[j].type;
	}
	if(peel_trace)	printf("peel_locus(): locus %s, sample_flag=%d, nn_all=%d, linktype=%d\n",locus<0?"QTL":marker[locus].name,sample_flag,nn_all,linktype);
	if(nn_all<2) return 0.0;
	count=freq+n_genetic_groups;
	for(j=0;j<n_genetic_groups;j++) {
		if(locus>=0 && marker[locus].count_flag[j]) for(i=0;i<nn_all;i++) count[j][i]=marker[locus].counts[j][i];
		else for(i=0;i<nn_all;i++) count[j][i]=1.0;
	}
	if(locus<0)	{
		for(i=0;i<nn_all;i++) {
			peel_alls[i]=i;
			for(j=0;j<n_genetic_groups;j++) freq[j][i]=tlocus[-1-locus].freq[j][i];
		}
		sample_freq=1;
	}
	if(sample_flag!=1) sample_freq=0;
	for(j=comp=0;comp<n_comp;comp++) {
		for(k=0;k<2;k++) {
			mem_block[k]=first_mem_block[k];
			mem_block[k]->ptr=0;
		}
		if(locus<0)	{
			rf=r_func[n_markers][comp];
			n_all=nn_all;
			lump=0;
		} else {
			rf=r_func[locus][comp];
			n_all=marker[locus].n_all1[comp];
			if(n_all<2)	{
				j+=comp_size[comp];
				continue;
			}
			for(i=0;i<nn_all;i++) {
				for(grp=0;grp<n_genetic_groups;grp++) freq[grp][i]=0.0;
				peel_alls[i]=n_all-1;
			}
			for(i=0;i<nn_all;i++) if((k=allele_trans[locus][comp][i])>=0) peel_alls[k]=i;
			for(grp=0;grp<n_genetic_groups;grp++)
			  for(i=0;i<nn_all;i++) freq[grp][peel_alls[i]]+=marker[locus].freq[grp][i];
			lump=0;
			if(n_all<nn_all) for(i=0;i<nn_all;i++) if(peel_alls[i]==n_all-1) lump|=(1<<i);
		}
		n_bits=num_bits(n_all);
		mask=(1<<n_bits)-1;
		if(locus<0) n_idx=n_all*n_all;
		else n_idx=1<<(n_bits+n_bits);
		n_peel_ops=0;
		for(i=0;i<comp_size[comp];i++) {
			i1=i+j;
			id_array[i1].flag=0;
			id_array[i1].allele[X_MAT]=id_array[i1].allele[X_PAT]= -1;
			k3=0;
			ids=id_array[i1].sire;
			idd=id_array[i1].dam;
			if(locus>=0) {
				marker[locus].m_flag[i1]=0;
				if(id_array[i1].pruned_flag[locus]) continue;
				if(ids && !id_array[ids-1].pruned_flag[locus]) k3=1;
			} else {
				if(id_array[i1].pruned_flag[n_markers]) continue;
				if(ids && !id_array[ids-1].pruned_flag[n_markers]) k3=1;
			}
			/* Set up transmission probs. */
			if(k3) {
				Mtp[0]=Mtp[1]=Ptp[0]=Ptp[1]=1.0;
				for(k=idx-1;k>=0;k--) {
					k1=perm[k];
					if(k1<0)	{
						if(id_array[i1].pruned_flag[n_markers] || id_array[ids-1].pruned_flag[n_markers] || !(tlocus[-1-k1].flag&MARKER_SAMPLED)) continue;
						k2=tlocus[-1-k1].seg[X_MAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(pos[X_MAT]-tlocus[-1-k1].pos[X_MAT])));
					} else {
						if(marker[k1].n_all1[comp]<2) continue;
						if(id_array[i1].pruned_flag[k1] || id_array[ids-1].pruned_flag[k1] || !(marker[k1].flag&MARKER_SAMPLED)) continue;
						k2=marker[k1].seg[X_MAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(pos[X_MAT]-marker[k1].pos[X_MAT])));
					}
					Mtp[k2]=1.0-theta;
					Mtp[1-k2]=theta;
					break;
				}
				for(k=idx-1;k>=0;k--) {
					k1=perm[k];
					if(k1<0)	{
						if(id_array[i1].pruned_flag[n_markers] || id_array[ids-1].pruned_flag[n_markers] || !(tlocus[-1-k1].flag&MARKER_SAMPLED)) continue;
						k2=tlocus[-1-k1].seg[X_PAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(pos[X_PAT]-tlocus[-1-k1].pos[X_PAT])));
					} else {
						if(marker[k1].n_all1[comp]<2) continue;
						if(id_array[i1].pruned_flag[k1] || id_array[ids-1].pruned_flag[k1] || !(marker[k1].flag&MARKER_SAMPLED)) continue;
						k2=marker[k1].seg[X_PAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(pos[X_PAT]-marker[k1].pos[X_PAT])));
					}
					Ptp[k2]=1.0-theta;
					Ptp[1-k2]=theta;
					break;
				}
				for(k=idx+1;k<n_loci;k++) {
					k1=perm[k];
					if(k1<0)	{
						if(id_array[i1].pruned_flag[n_markers] || id_array[ids-1].pruned_flag[n_markers] || !(tlocus[-1-k1].flag&MARKER_SAMPLED)) continue;
						k2=tlocus[-1-k1].seg[X_MAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(tlocus[-1-k1].pos[X_MAT]-pos[X_MAT])));
					} else {
						if(marker[k1].n_all1[comp]<2) continue;
						if(id_array[i1].pruned_flag[k1] || id_array[ids-1].pruned_flag[k1] || !(marker[k1].flag&MARKER_SAMPLED)) continue;
						k2=marker[k1].seg[X_MAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(marker[k1].pos[X_MAT]-pos[X_MAT])));
					}
					Mtp[k2]*=1.0-theta;
					Mtp[1-k2]*=theta;
					break;
				}
				for(k=idx+1;k<n_loci;k++) {
					k1=perm[k];
					if(k1<0)	{
						if(id_array[i1].pruned_flag[n_markers] || id_array[ids-1].pruned_flag[n_markers] || !(tlocus[-1-k1].flag&MARKER_SAMPLED)) continue;
						k2=tlocus[-1-k1].seg[X_PAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(tlocus[-1-k1].pos[X_PAT]-pos[X_PAT])));
					} else {
						if(marker[k1].n_all1[comp]<2) continue;
						if(id_array[i1].pruned_flag[k1] || id_array[ids-1].pruned_flag[k1] || !(marker[k1].flag&MARKER_SAMPLED)) continue;
						k2=marker[k1].seg[X_PAT][i1];
						if(k2<0) continue;
						theta=.5*(1.0-exp(-0.02*(marker[k1].pos[X_PAT]-pos[X_PAT])));
					}
					Ptp[k2]*=1.0-theta;
					Ptp[1-k2]*=theta;
					break;
				}
				z=Mtp[0]+Mtp[1];
				z1=Ptp[0]+Ptp[1];
				for(k=0;k<2;k++) {
					Mtp[k]/=z;
					Ptp[k]/=z1;
					id_array[i1].tpp[X_MAT][k]=Mtp[k];
					id_array[i1].tpp[X_PAT][k]=Ptp[k];
				}
				id_array[i1].tp[X_MM_PM]=Mtp[X_MAT]*Ptp[X_MAT];
				id_array[i1].tp[X_MM_PP]=Mtp[X_MAT]*Ptp[X_PAT];
				id_array[i1].tp[X_MP_PM]=Mtp[X_PAT]*Ptp[X_MAT];
				id_array[i1].tp[X_MP_PP]=Mtp[X_PAT]*Ptp[X_PAT];
			} else {	
				for(k=0;k<2;k++) id_array[i1].tpp[X_MAT][k]=id_array[i1].tpp[X_PAT][k]=0.5;
				for(k=0;k<4;k++) id_array[i1].tp[k]=0.25;
			}
			id_array[i1].rfp= -1;
			if(locus>=0) {
				for(k1=0;k1<2;k1++) {
					a=req_set[k1][locus][i1];
					if(a) {
						for(k=0;k<n_all;k++) if(a&(1<<k)) break;
						id_array[i1].lumped[k1]=k;
					} else id_array[i1].lumped[k1]=0;
				}
				id_array[i1].ngens=id_array[i1].nhaps[0]=id_array[i1].nhaps[1]=0;
				k1=k1a=k2=k3=0;
				for(b=b1=0,k=0;k<n_all;k++) {
					a=a_set[i1][k];
					if(!a) continue;
					k1++;
					k1a=k;
					b|=a;
					b1|=(1<<k);
					while(a)	{
						if(a&1) k3++;
						a>>=1;
					}
				}
				id_array[i1].nhaps[X_MAT]=k1;
				if(k1==1) id_array[i1].allele[X_MAT]=k1a+1;
				id_array[i1].temp[X_PAT]=b;
				id_array[i1].temp[X_MAT]=b1;
				k1a=k2a=0;
				while(b)	{
					if(b&1) {
						k2++;
						k2a=k1a;
					}
					k1a++;
					b>>=1;
				}
				id_array[i1].nhaps[X_PAT]=k2;
				if(k2==1) id_array[i1].allele[X_PAT]=k2a+1;
				id_array[i1].ngens=k3;
				if(k3==1 && id_array[i1].allele[X_MAT]==id_array[i1].allele[X_PAT] && id_array[i1].nkids)	{
					marker[locus].m_flag[i1]=1;
				}
				if(ids) {
					b=req_set[X_PAT][locus][i1];
					b1=id_array[i1].temp[X_PAT];
					if(b&b1) b1|=b;
					b=id_array[ids-1].temp[X_PAT];
					trn[X_PAT]=(b1&b)?1:0;
					b=id_array[ids-1].temp[X_MAT];
					trn[X_MAT]=(b1&b)?1:0;
					if(!trn[0] || !trn[1]) marker[locus].m_flag[i1]|=2;
				}
				if(idd) {
					b=req_set[X_MAT][locus][i1];
					b1=id_array[i1].temp[X_MAT];
					if(b&b1) b1|=b;
					b=id_array[idd-1].temp[X_PAT];
					trn[X_PAT]=(b1&b)?1:0;
					b=id_array[idd-1].temp[X_MAT];
					trn[X_MAT]=(b1&b)?1:0;
					if(!trn[0] || !trn[1]) marker[locus].m_flag[i1]|=4; 
				}
			}
		}
		if(locus>=0 && marker[locus].mterm) pen=&penetrance;
		n_rf= -1;
		sample_flag&=~OP_SAMPLING;
		if(peel_trace) printf("Peeling component %d\n",comp+1);
		/* Go through all operations in peeling sequence */
		if(locus<0) pp=peelseq_head[n_markers]+comp;
		else pp=peelseq_head[locus]+comp;
		while(pp->type) {
			if(pp->type==PEEL_SIMPLE) {
				simple_em=pp->ptr.simple;
				if(sample_flag && simple_em->pivot) 
				  peel_list[n_peel_ops++]=pp;
				if(locus<0)	{
					if(simple_em->out_index>n_rf)	{
						n_rf=simple_em->out_index;
						rf[n_rf].p=0;
					}
					z=loki_trait_simple_peelop(simple_em,locus,sample_flag);
					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Internal error - shouldn't be here\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					like+=z;
				} else like+=loki_simple_peelop(simple_em,locus,sample_flag,*pen);
				pp= &simple_em->next;
			} else {
				complex_em=pp->ptr.complex;
				if(sample_flag && (complex_em->n_involved-complex_em->n_peel))
				  peel_list[n_peel_ops++]=pp;
				if(locus<0)	{
					if(complex_em->out_index>n_rf) {
						n_rf=complex_em->out_index;
						rf[n_rf].p=0;
					}
					z=loki_trait_complex_peelop(complex_em,locus,sample_flag);
					if(z== -DBL_MAX) {
						if(sample_flag&1) ABT_FUNC("Internal error - shouldn't be here\n");
						/* Error return on likelihood (not sampling) run.  Clean up and return error */
						return z;
					}
					like+=z;
				} else like+=loki_complex_peelop(complex_em,locus,sample_flag,*pen);
				pp= &complex_em->next;
			}
		}
		/* If we're sampling, do them again in reverse */
		if(sample_flag) {
			sample_flag|=OP_SAMPLING;
			for(i=n_peel_ops-1;i>=0;i--) {
				pp=peel_list[i];
				if(pp->type==PEEL_SIMPLE) {
					simple_em=pp->ptr.simple;
					if(simple_em->pivot== -2) {
						k1=simple_em->out_index;
					} else {
						if(locus<0) (void)loki_trait_simple_sample(simple_em,locus,sample_flag);
						else (void)loki_simple_sample(simple_em,locus,*pen);
					}
				} else {
					complex_em=pp->ptr.complex;	
					k=0;
					if(complex_em->n_peel==1) {
						k1=complex_em->involved[0];
						if(locus>=0) {
							if(k1>0)	{
								if(id_array[k1-1].nhaps[X_MAT]==1) {
									k=1;
									id_array[k1-1].flag|=SAMPLED_MAT;
								}
							} else {
								if(id_array[-k1-1].nhaps[X_PAT]==1)	{
									k=1;
									id_array[-k1-1].flag|=SAMPLED_PAT;
								}
							}
						}
					}
					if(!k) {
						if(locus<0) (void)loki_trait_complex_peelop(complex_em,locus,sample_flag);
						else (void)loki_complex_peelop(complex_em,locus,sample_flag,*pen);
					}
				}
			}
			/* Get allele counts for frequency update */
			if(sample_freq) for(i=0;i<comp_size[comp];i++) {
				i1=i+j;
				if(locus>=0) {
					if(id_array[i1].pruned_flag[locus]) continue;
					ids=id_array[i1].sire;
					if(ids && !id_array[ids-1].pruned_flag[locus]) continue;
					grp=id_array[i1].group-1;
					if(grp<0 || grp>=n_genetic_groups) ABT_FUNC("Internal error - invalid genetic group for founder\n");
					for(k1=0;k1<2;k1++) {
						allele=id_array[i1].allele[k1]-1;
						a=req_set[k1][locus][i1];
						if(!(a&(1<<allele))) a=1<<allele;
						z=0.0;
						for(k=0;k<n_all;k++) if(a&(1<<k)) {
							k2=peel_alls[k];
							if(lump&(1<<k2)) {
								for(k3=0;k3<n_all;k3++) if(lump&(1<<k3)) z+=marker[locus].freq[grp][k3];
							} else z+=marker[locus].freq[grp][k2];
						}
						for(k=0;k<n_all;k++) if(a&(1<<k)) {
							k2=peel_alls[k];
							if(lump&(1<<k2)) {
								for(k3=0;k3<n_all;k3++) if(lump&(1<<k3)) count[grp][k3]+=marker[locus].freq[grp][k3]/z;
							} else count[grp][k2]+=marker[locus].freq[grp][k2]/z;
						}
					}
				} else {
					if(id_array[i1].pruned_flag[n_markers]) continue;
					ids=id_array[i1].sire;
					if(ids && !id_array[ids-1].pruned_flag[n_markers]) continue;
					grp=id_array[i1].group-1;
					if(grp<0 || grp>=n_genetic_groups) ABT_FUNC("Internal error - invalid genetic group for founder\n");
					for(k1=0;k1<2;k1++) {
						allele=id_array[i1].allele[k1]-1;
						count[grp][allele]+=1.0;
					}
				}
			}
			/* Correct residuals, get segregation pattern */
			eff=0;
			mtype=0;
			if(locus<0 || marker[locus].mterm) {
				mtype=model.var.type;
				if(locus<0) eff=tlocus[-1-locus].eff;
				else if(marker[locus].mterm) eff=marker[locus].mterm->eff;
			}
			for(i=0;i<comp_size[comp];i++) {
				i1=i+j;
				if(locus>=0) {
					if(id_array[i1].pruned_flag[locus]) continue;
				} else if(id_array[i1].pruned_flag[n_markers]) continue;
				k=id_array[i1].allele[X_MAT];
				k1=id_array[i1].allele[X_PAT];
				if(k>k1) k=k*(k-1)/2+k1;
				else k=k1*(k1-1)/2+k;
				if(locus<0 || marker[locus].mterm) {
					/* Correct residuals for new genotypes */
					if(eff && id_array[i1].res) {
						nrec=id_array[i1].n_rec;
						if(!nrec || (mtype&ST_CONSTANT)) nrec=1;
						if(locus<0) {
							for(rec=0;rec<nrec;rec++) {
								z=id_array[i1].res[rec];
								if(tlocus[-1-locus].flag&MARKER_SAMPLED) {
									k1=tlocus[-1-locus].gt[i1]-1;
									if(k1) z+=eff[k1-1];
								}
								if(k>1) z-=eff[k-2];
								id_array[i1].res[rec]=z;
							}
						} else {
							for(rec=0;rec<nrec;rec++) {
								z=id_array[i1].res[rec];
								if(marker[locus].flag&MARKER_SAMPLED) {
									k1=marker[locus].gt[i1]-1;
									if(k1) z+=eff[k1-1];
								}
								if(k>1) z-=eff[k-2];
								id_array[i1].res[rec]=z;
							}
						}
					}
				}
				if(locus<0) tlocus[-1-locus].gt[i1]=k;
				else marker[locus].gt[i1]=k;
				if(locus<0) {
					idd=id_array[i1].dam;
					if(idd && !id_array[idd-1].pruned_flag[n_markers])	{
						k=id_array[i1].allele[X_MAT];
						if(k<1 || k>n_all) {
							ABT_FUNC("Internal error - bad sampled genotype\n");
						}
						for(k1=0;k1<2;k1++)
						  trn[k1]=(k==id_array[idd-1].allele[k1])?1:0;
						if(!trn[0] && !trn[1]) {
							ABT_FUNC("Internal error - bad configuration?\n");
						}
						if(trn[0] && trn[1])	{
							if(si_mode)	{
								tpp=id_array[i1].tpp[X_MAT];
								z=ranf()*(tpp[0]+tpp[1]);
								seg[X_MAT][i1]=(z<=tpp[0])?0:1;
							} else seg[X_MAT][i1]= -2;
						} else if(trn[0]) seg[X_MAT][i1]=0;
						else seg[X_MAT][i1]=1;
					} else seg[X_MAT][i1]= -1;
					ids=id_array[i1].sire;
					if(ids && !id_array[ids-1].pruned_flag[n_markers])	{
						k=id_array[i1].allele[X_PAT];
						if(k<1 || k>n_all) ABT_FUNC("Internal error - bad sampled genotype\n");
						for(k1=0;k1<2;k1++)
						  trn[k1]=(k==id_array[ids-1].allele[k1])?1:0;
						if(!trn[0] && !trn[1]) {
							ABT_FUNC("Internal error - bad configuration?\n");
						}
						if(trn[0] && trn[1])	{
							if(si_mode)	{
								tpp=id_array[i1].tpp[X_PAT];
								z=ranf()*(tpp[0]+tpp[1]);
								seg[X_PAT][i1]=(z<=tpp[0])?0:1;
							} else seg[X_PAT][i1]= -2;
						} else if(trn[0]) seg[X_PAT][i1]=0;
						else seg[X_PAT][i1]=1;
					} else seg[X_PAT][i1]= -1;
				} else {
					idd=id_array[i1].dam;
					if(idd && !id_array[idd-1].pruned_flag[locus]) {
						c=req_set[X_MAT][locus][i1];
						k=id_array[i1].allele[X_MAT]-1;
						if(k<0 || k>=n_all) ABT_FUNC("Internal error - bad sampled genotype\n");
						a=1<<k;
						if(a&c) a=c;
						for(k1=0;k1<2;k1++) {
							k=id_array[idd-1].allele[k1]-1;
							b=1<<k;
							trn[k1]=(a&b)?1:0;
						}
						if(!trn[0] && !trn[1]) {
							ABT_FUNC("Internal error - bad configuration?\n");
						}
						if(trn[0] && trn[1])	{
							if(si_mode)	{
								tpp=id_array[i1].tpp[X_MAT];
								z=ranf()*(tpp[0]+tpp[1]);
								seg[X_MAT][i1]=(z<=tpp[0])?0:1;
							} else seg[X_MAT][i1]= -2;
						} else {
							seg[X_MAT][i1]=trn[0]?0:1;
						}
					} else seg[X_MAT][i1]= -1;
					ids=id_array[i1].sire;
					if(ids && !id_array[ids-1].pruned_flag[locus]) {
						c=req_set[X_PAT][locus][i1];
						k=id_array[i1].allele[X_PAT]-1;
						if(k<0 || k>=n_all) ABT_FUNC("Internal error - bad sampled genotype\n");
						a=1<<k;
						if(a&c) a=c;
						for(k1=0;k1<2;k1++) {
							k=id_array[ids-1].allele[k1]-1;
							b=1<<k;
							trn[k1]=(a&b)?1:0;
						}
						if(!trn[0] && !trn[1]) {
							ABT_FUNC("Internal error - bad configuration?\n");
						}
						if(trn[0] && trn[1]) {
							if(si_mode)	{
								tpp=id_array[i1].tpp[X_PAT];
								z=ranf()*(tpp[0]+tpp[1]);
								seg[X_PAT][i1]=(z<=tpp[0])?0:1;
							} else seg[X_PAT][i1]= -2;
						} else {
							seg[X_PAT][i1]=trn[0]?0:1;
						}
					} else seg[X_PAT][i1]= -1;
				}
			}
		}
		j+=comp_size[comp];
	}
   if(sample_flag && sample_freq) for(grp=0;grp<n_genetic_groups;grp++) {
		z=z1=0.0;
		if(locus<0) for(i=0;i<nn_all;i++) z1+=count[grp][i];
		else for(i=0;i<nn_all;i++)	{
			if(marker[locus].freq_set[grp][i]!=1) z1+=count[grp][i];
			else z+=marker[locus].freq[grp][i];
		}
		for(i=0;i<nn_all-1;i++) if(locus<0 || marker[locus].freq_set[grp][i]!=1) {
			z1-=count[grp][i];
			freq[grp][i]=(1.0-z)*genbet(count[grp][i],z1);
			z+=freq[grp][i];
		}
		if(locus<0 || marker[locus].freq_set[grp][i]!=1) freq[grp][i]=1.0-z;
		if(locus<0)	{
			for(i=0;i<nn_all;i++) tlocus[-1-locus].freq[grp][i]=freq[grp][i];
		} else for(i=0;i<nn_all;i++) marker[locus].freq[grp][i]=freq[grp][i];
	}
	if(sample_flag) {
		if(locus<0) tlocus[-1-locus].flag|=MARKER_SAMPLED;
		else marker[locus].flag|=MARKER_SAMPLED|RFMASK_OK;
	}
	return like;
}
