/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - Rockefeller University                         *
 *                                                                          *
 *                       October 1997                                       *
 *                                                                          *
 * loki_sample.c:                                                           *
 *                                                                          *
 * Main sampling loop + small sample routines                               *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/times.h>
#include <time.h>
#include <sys/wait.h>

#include "ranlib.h"
#include "version.h"
#include "utils.h"
#include "loki.h"
#include "loki_peel.h"
#include "loki_ibd.h"
#include "loki_dump.h"
#include "loki_tlmoves.h"
#include "sample_cens.h"
#include "handle_res.h"
#include "sample_effects.h"
#include "sample_rand.h"
#include "print_data.h"
#include "calc_var_locus.h"
#include "loki_output.h"
#include "kinship.h"
#include "genedrop.h"
#include "sample_nu.h"

double tau_beta=2.0,tau,lm_ratio,l_prob;
int no_overdominant,tau_mode,l_cycles,analysis=DEFAULT_ANALYSIS;
int max_linked=2,max_linked_set=1;

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "SampleLoop"
void SampleLoop(void)
{
	int lp,i,j,j1,j2,k,k1,k2,flag,comp,*perm=0,*naffect=0,**affs=0,dumped=0,ibdflag,*gt,*ntl_linked=0,*ntl_linked1=0,lp1;
	FILE *fptr=0,*fibd=0,*ffreq=0,*fpos=0,*fmpos=0;
 	double z,z1,*ss=0,*ss2=0,**pairs=0,*trpos=0,*trpos1=0;
	struct output_gen *og;
	double addlog(double x1,double x2);
	char *fname=0,*posname=0,*mposname=0;
	int ct1[3],ct2[3];
	long old_pos=-1;
	
	if(Output_Phen) {
		Print_Data(Output_Phen);
		free(Output_Phen);
	}
	while(Output_Gen)	{
		Print_Genotypes(Output_Gen);
		og=Output_Gen->next;
		free(Output_Gen->file);
		free(Output_Gen);
		Output_Gen=og;
	}
	ibdflag=SetupIBD();
	if(ibdflag) analysis=(IBD_ANALYSIS|ESTIMATE_IBD);
	/* if si_mode=1, Sample SI's where genotypes do not completely determine segregation pattern */
	printf("Analysis = %d\n",analysis);
	sort_sex=0;
	TL_Alloc();
	lp=lp1=0;
	z=RES_PRIOR_V0*.5;
	res_prior_konst=log(z)*z-lgamma(z)+log(RES_PRIOR_S0)*z;
	if(read_dump_flag) {
		fputs("Retrieving program state: ",stdout);
		fflush(stdout);
		j=read_dump(&lp,&lp1,&old_pos);
		if(j<0) {
			fputs("FAILED\n",stdout);
			ABT_FUNC("Aborting\n");
		} else fputs("OK\n",stdout);
		do j=wait(&i); while(j>0); /* Wait for child processes to finish */
	} else {
		if(!tlocus) start_tloci=0;
		for(i=0;i<start_tloci;i++)	{
			k=get_new_traitlocus(2);
			if(k<0) ABT_FUNC("Internal error - Couldn't get trait locus\n");
			for(j=0;j<n_genetic_groups;j++) tlocus[k].freq[j][0]=tlocus[k].freq[j][1]=0.5;
			tlocus[k].eff[0]=tlocus[k].eff[1]=0.0;
			tlocus[k].link_group= -1;
			tlocus[k].pos[0]=tlocus[k].pos[1]=0.0;
			tlocus[k].flag=TL_UNLINKED;
			k1=-k-1;
			(void)peel_locus(&k1,0,1,1);
		}
		/* Get initial genotype samples for all markers */
		if(!(analysis&NULL_ANALYSIS)) for(k=0;k<n_links;k++) Sample_LinkageGroup(k);
		/* Handle weights for t model */
		if(use_student_t) {
			init_sample_nu();
			sample_nu();
		}
	}
	if(si_mode) flag=1;
	else flag=0;
	if(!(analysis&NULL_ANALYSIS)) {
		if(analysis&IBD_ANALYSIS) {
			if(!(flag&1)) {
				sample_segs();
				flag|=1;
			}
			for(k=0;k<n_markers;k++) for(comp=0;comp<n_comp;comp++) pass_founder_genes(k,comp);
		}
	}
	if(ibdflag) fname=make_file_name("ibd");
	else {
		if(!Outputfile) Outputfile=make_file_name("out");
		if(!(analysis&(IBD_ANALYSIS|IBS_ANALYSIS))) posname=make_file_name("pos");
	}
	for(i=0;i<n_markers;i++) if(marker[i].pos_set==2) break;
	if(i<n_markers) mposname=make_file_name("mpos");
	if(append_output_flag) {
		if(ibdflag) if(!(fibd=fopen(fname,"a"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,fname);
		if(mposname) if(!(fmpos=fopen(mposname,"a"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,mposname);
		if(Freqfile) if(!(ffreq=fopen(Freqfile,"a"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,Freqfile);
		if(!ibdflag) {
			if(!(fptr=fopen(Outputfile,"a"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,Outputfile);
			if(posname) if(!(fpos=fopen(posname,"r+"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,posname);
		}
	} else {
		j=syst_var[SYST_BACKUPS].flag?syst_var[SYST_BACKUPS].data.value:1;
		if(j) {
			i=mkbackup(Outputfile,j);
			if(ibdflag) i=mkbackup(fname,j);
		}
		if(ibdflag) if(!(fibd=fopen(fname,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,fname);
		if(mposname) if(!(fmpos=fopen(mposname,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,mposname);
		if(Freqfile) if(!(ffreq=fopen(Freqfile,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,Freqfile);
		if(!ibdflag) {
			if(!(fptr=fopen(Outputfile,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,Outputfile);
			if(posname) if(!(fpos=fopen(posname,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for output\n",FUNC_NAME,posname);
			OutputHeader(fptr);
		}
		if(ffreq) OutputFreqHeader(ffreq);
	}
	if(posname) free(posname);
	if(mposname) free(mposname);
	if(fname) free(fname);
	OutputHeader(stdout);
	if(analysis&(IBD_ANALYSIS|IBS_ANALYSIS)) {
		if(n_markers) {
			k=2;
			if(!(ss=malloc(sizeof(double)*n_markers*k))) ABT_FUNC(MMsg);
			if(!(perm=malloc(sizeof(int)*n_markers))) ABT_FUNC(MMsg);
			ss2=ss+n_markers;
		}
		for(i=0;i<n_markers;i++) ss[i]=ss2[i]=0.0;
		if(!(naffect=malloc(sizeof(void *)*n_comp))) ABT_FUNC(MMsg);
		if(!(affs=malloc(sizeof(void *)*n_comp))) ABT_FUNC(MMsg);
		if(!(pairs=malloc(sizeof(void *)*n_comp))) ABT_FUNC(MMsg);
		for(i=comp=0;comp<n_comp;comp++) {
			naffect[comp]=0;
			for(j=0;j<comp_size[comp];j++) if(id_array[i+j].affected==2) naffect[comp]++;
			if(naffect[comp]) {
				j=naffect[comp]*(naffect[comp]+1)/2;
				if(!(affs[comp]=malloc(sizeof(int)*naffect[comp]))) ABT_FUNC(MMsg);
				if(!(pairs[comp]=malloc(sizeof(double)*j))) ABT_FUNC(MMsg);
				k=0;
				for(j=0;j<comp_size[comp];j++) if(id_array[i+j].affected==2) affs[comp][k++]=i+j;
				for(k1=j=0;j<naffect[comp];j++) for(k=0;k<=j;k++) {
					pairs[comp][k1++]=kinship(affs[comp][j]+1,affs[comp][k]+1);
				}
			} else {
				affs[comp]=0;
				pairs[comp]=0;
			}
			i+=comp_size[comp];
		}
	} else if(tlocus && max_tloci) {
		if(!(ntl_linked=malloc(sizeof(int)*(n_links+1)*2))) ABT_FUNC(MMsg);
		ntl_linked1=ntl_linked+n_links+1;
		for(i=0;i<=n_links;i++) ntl_linked1[i]=-1;
		if(!(trpos=malloc(sizeof(double)*(1+sex_map)*2*max_tloci))) ABT_FUNC(MMsg);
		trpos1=trpos+(1+sex_map)*max_tloci;
		if(read_dump_flag) {
			for(k=0;k<=n_links;k++) ntl_linked1[k]=0;
			for(k1=0;k1<n_tloci;k1++) if(tlocus[k1].flag && !(tlocus[k1].flag&TL_LINKED)) ntl_linked1[0]++;
			for(k2=k=0;k<n_links;k++) {
				for(k1=0;k1<n_tloci;k1++) if(tlocus[k1].flag && (tlocus[k1].flag&TL_LINKED) && tlocus[k1].link_group==k) {
				ntl_linked1[k+1]++;
					trpos1[k2++]=tlocus[k1].pos[0];
					if(sex_map) trpos1[k2++]=tlocus[k1].pos[1];
				}
			}
		}
	}
	catch_sigs=1;
	for(++lp;!sig_caught && (!num_iter || lp<=num_iter);lp++) {
#ifdef USE_DMALLOC
		if(dmalloc_verify(0)==DMALLOC_ERROR) {
			fprintf(stderr,"[%s:%d] %s(): Error returned from dmalloc_verify().\nAttempting to abort nicely.\n",__FILE__,__LINE__,FUNC_NAME);
			break;
		}
#endif
		if(!(analysis&NULL_ANALYSIS)) {
			if(ranf()<lm_ratio) {
				/* M sample */
				if(!(flag&1)) {
					sample_segs();
					flag|=1;
				}
				if(!(flag&2)) {
					for(k=0;k<n_markers;k++) {
						for(comp=0;comp<n_comp;comp++) {
							pass_founder_genes(k,comp);
							marker[k].lk_store[comp]=seg_pen(k,comp,&i,0);
							if(i) ABT_FUNC("Illegal segregation pattern\n");
						}
					}
 					flag|=2;
				}
				for(k=0;k<n_links;k++) meiosis_scan(k);
			} else { /* L sample */
				if(lm_ratio>0.0 && !si_mode && (flag&1)) {
					for(k=0;k<n_markers;k++) {
						for(comp=0;comp<n_comp;comp++) {
  							if(!(flag&2)) pass_founder_genes(k,comp);
							(void)seg_pen(k,comp,&i,1);
							if(i) {
								fprintf(stderr,"%d %d %d\n",k,comp,i);
								ABT_FUNC("Illegal segregation pattern\n");
							}
						}
					}
					flag&=~1;
				}
				for(k=0;k<n_links;k++) Sample_LinkageGroup(k);
				if(si_mode) flag&=~2;
				else flag&=~3;
			}
		}
		if(analysis&ESTIMATE_IBD) {
 			Handle_IBD();
			if(sample_freq[0] && lp>=sample_from[0] && !(lp%sample_freq[0]))
			  Output_Sample_IBD(fibd,lp);
			if(sample_freq[1] && lp>=sample_from[1] && !(lp%sample_freq[1]))
			  printf("At: %d\n",lp);
		} else if(analysis&(IBD_ANALYSIS|IBS_ANALYSIS)) {
			if(analysis&IBD_ANALYSIS) {
				if(analysis&NULL_ANALYSIS) genedrop(0);
				for(k=0;k<n_markers;k++) {
					z=0.0;
					k2=0;
					for(comp=0;comp<n_comp;comp++) {
						if(naffect[comp]>1) {
							if(!(flag&2)) {
								pass_founder_genes(k,comp);
								flag|=2;
							}
							z+=score_ibd(k,&k1,comp,naffect[comp],affs[comp],pairs[comp]);
							k2+=k1;
						}
					}
					if(k2) z=sqrt(z/(double)k2);
					ss[k]=z;
					if(analysis&NULL_ANALYSIS) break;
				}
			} else if(analysis&IBS_ANALYSIS) {
				for(k=0;k<n_markers;k++) {
					if(analysis&NULL_ANALYSIS) drop_genotypes(k);
					gt=marker[k].gt;
					for(j=0;j<3;j++) ct1[j]=ct2[j]=0;
					for(k1=k2=j=0;j<ped_size;j++) if(id_array[j].affected) {
						if(id_array[j].affected==1) {
							ct1[gt[j]-1]++;
							k1++;
						} else if(id_array[j].affected==2) {
							ct2[gt[j]-1]++;
							k2++;
						}
					}
					z=k1?((double)ct1[0]+.5*(double)ct1[1])/(double)k1:0.0;
					z1=k2?((double)ct2[0]+.5*(double)ct2[1])/(double)k2:0.0;
					ss[k]=z;
					ss2[k]=z1;
				}
			}
			if(sample_freq[0] && lp>=sample_from[0] && !(lp%sample_freq[0])) {
				fprintf(fptr,"%d",lp);
				for(k=0;k<n_markers;k++) fprintf(fptr," %g %g %g",ss[k],ss2[k],ss[k]-ss2[k]);
				fputc('\n',fptr);
			}
			if(sample_freq[1] && lp>=sample_from[1] && !(lp%sample_freq[1])) {
				fprintf(stdout,"%d",lp);
				for(k=0;k<n_markers;k++) fprintf(stdout," %g %g %g",ss[k],ss2[k],ss[k]-ss2[k]);
				fputc('\n',stdout);
			}
		} else {
			if(tlocus) TL_Birth_Death();
			for(k=0;k<n_tloci;k++) if(tlocus[k].flag) {
				Sample_TL_Position(k);
			}
			if(use_student_t) sample_nu();
 			if(censored_flag) Sample_Censored(); 
			sample_effects();
 			if(res_var_set!=1) Sample_ResVar();
 			if(n_random) sample_rand();
 			if(polygenic_flag) sample_additive_var();
			switch(tau_mode) {
			 case 0:
			 case 1:
				tau=tau_beta;
				break;
			 case 2:
				tau=residual_var*tau_beta;
				break;
			}
			if(debug_level&4) fputc('\n',stdout);
			k=0;
			if(sample_freq[0] && lp>=sample_from[0] && !(lp%sample_freq[0])) {
				for(i=0;i<n_tloci;i++) if(tlocus[i].flag)
				  tlocus[i].variance=calc_var_locus(-1-i);
				OutputSample(fptr,lp);
				if(Freqfile) OutputFreq(ffreq,lp);
				k=1;
			}
			if(sample_freq[1] && lp>=sample_from[1] && !(lp%sample_freq[1])) {
				if(!k) for(i=0;i<n_tloci;i++) if(tlocus[i].flag)
				  tlocus[i].variance=calc_var_locus(-1-i);
				OutputSample(stdout,lp);
			}
			if(dump_freq && lp && !(lp%dump_freq)) {
				write_dump(lp,lp1,old_pos);
				dumped=1;
			} else dumped=0;
			if(fmpos && lp>=sample_from[0]) {
				if(lp==1) {
					for(k=i=0;i<n_markers;i++) if(marker[i].pos_set==2) {
						if(k++) fputc(' ',fmpos);
						fprintf(fmpos,"%s",marker[i].name);
						if(marker[i].index) fprintf(fmpos,"(%d)",marker[i].index);
					}
					fputc('\n',fmpos);
				}
				fprintf(fmpos,"%d ",lp);
				for(i=0;i<n_markers;i++) if(marker[i].pos_set==2) {
					fprintf(fmpos,"%g ",marker[i].pos[0]);
					if(sex_map) fprintf(fmpos,"%g ",marker[i].pos[1]);
				}
				fputc('\n',fmpos);
				fflush(fmpos);
			}
			if(ntl_linked && fpos && lp>=sample_from[0]) {
				for(k=0;k<=n_links;k++) ntl_linked[k]=0;
				for(k1=0;k1<n_tloci;k1++) if(tlocus[k1].flag && !(tlocus[k1].flag&TL_LINKED)) ntl_linked[0]++;
				for(k2=k=0;k<n_links;k++) {
					for(k1=0;k1<n_tloci;k1++) if(tlocus[k1].flag && (tlocus[k1].flag&TL_LINKED) && tlocus[k1].link_group==k) {
						ntl_linked[k+1]++;
						trpos[k2++]=tlocus[k1].pos[0];
						if(sex_map) trpos[k2++]=tlocus[k1].pos[1];
					}
				}
				for(j=k=0;k<=n_links;k++) if(ntl_linked[k]!=ntl_linked1[k]) break;
				if(k==n_links+1) {
					for(k=0;k<k2;k++) if(trpos[k]!=trpos1[k]) break;
					if(k==k2) j=1;
				}
				if(old_pos>=0) fseek(fpos,old_pos,SEEK_SET);
				else {
					fseek(fpos,0,SEEK_END);
					old_pos=ftell(fpos);
				}
				if(lp1) fprintf(fpos,":%d\n",lp-lp1);
				if(!j) {
					for(j1=k2=k=0;k<=n_links;k++) {
						ntl_linked1[k]=ntl_linked[k];
						if(ntl_linked[k]) {
							if(j1++) fputc(' ',fpos);
							fprintf(fpos,"%d %d",k,ntl_linked[k]);
							if(k) for(k1=0;k1<ntl_linked[k];k1++) {
								for(j2=0;j2<1+sex_map;j2++) {
									trpos1[k2]=trpos[k2];
									fprintf(fpos," %g",trpos[k2++]);
								}
							}
						}
					}
					lp1=lp;
					old_pos=ftell(fpos);
					fputs(":1\n",fpos);
				}
				fflush(fpos);
			}
		}
	}
	if(!dumped && !(analysis&(IBD_ANALYSIS|IBS_ANALYSIS))) write_dump(lp-1,lp1,old_pos);
	if(fptr) (void)fclose(fptr);
	if(fibd) (void)fclose(fibd);
	if(fpos) (void)fclose(fpos);
	if(ffreq) (void)fclose(ffreq);
	if(ss) free(ss);
	if(perm) free(perm);
	if(naffect) free(naffect);
	if(affs) free(affs);
	if(pairs) free(pairs);
	if(ntl_linked) free(ntl_linked);
	if(trpos) free(trpos);
}
