%{
/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       July 1997                                          *
 *                                                                          *
 * param_parse.y:                                                           *
 *                                                                          *
 * yacc source for parameter file parser.                                   *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "utils.h"
#include "loki_scan.h"
#include "shared_peel.h"
#include "loki.h"
	
extern FILE *yyin;
double residual_var,residual_var_limit,additive_var,additive_var_limit,grand_mean,tloci_mean;
int res_var_set,add_var_set,grand_mean_set,num_iter,max_tloci=16,min_tloci,sex_map;
int tloci_mean_set,debug_level,start_tloci=-1,dump_freq,output_type=DEFAULT_OUTPUT_TYPE;
int sample_from[2]={0,0};
int sample_freq[2]={1,1};
static int start_flag=1;
static struct Marker *freq_marker;
static int check_variance( double);
static void set_position(struct lk_variable *, double, double);
static struct lk_variable *find_var( char *, int, int);
static struct Marker *check_marker(struct lk_variable *);
static int find_allele( char *, struct Marker *,int, int);
static int find_group( char *,int, int);
static void set_output_gen(char *,char *);
static struct IBD_List *add_ibd_list(double,struct IBD_List *);
static void set_freq(struct Marker *, double,int);
static void set_map_range( char *,double,double, int);
static void set_tloci(int,int);	
static void set_ibd_list( char *,struct IBD_List *);
static void set_output(struct lk_variable *);
static void set_group_order( int);
static void set_analyze(char *);
static struct Variable *group_var;
static int group_ptr,*group_order,group_counter,freq_allele,c_flag;
extern void yyerror(char *s),print_scan_err(char *fmt, ...);
extern int yyparse(void),yylex(void),lineno,lineno1,tokenpos;
extern char *yytext,linebuf[];
char *Output_Phen,*Outputfile,*Dumpfile,*Freqfile,*Haplofile;
struct output_gen *Output_Gen;

static int max_scan_errors=30;
static int scan_warn_n,max_scan_warnings=30;
int scan_error_n,output_haplo;

struct id_data syst_var[NUM_SYSTEM_VAR];

%}

%union  {
	char *string;
	int value;
	double rvalue;
	struct IBD_List *rlist;
	struct lk_variable *lk_var;
}

%token RESIDUAL GENETIC VARIANCE POSITION FREQUENCY 
%token START MEAN ITERATIONS SAMPLE FROM OUTPUT MAP TOTAL
%token SEED SFILE SEEDFILE TRAIT LOCI SET SYSTEM_VAR
%token ESTIMATE IBD GROUP ORDER MALE FEMALE LIMIT
%token PHENO GENO COUNTS DUMP TYPE ANALYZE NORMAL STUDENT_T
%token HAPLO

%token <string> STRING
%token <value> INTEGER SYSTEM_VAR
%token <rvalue> REAL
%type <rvalue> rnum
%type <rlist> ibdlist
%type <lk_var> lkvar
%type <value> allele,group

%%

parmfile: {lineno1=lineno;} command1 
       | error
       | parmfile {lineno1=lineno;} command1
       | parmfile error
       ;

command1: command
       | command_a
       | START {start_flag=2;} command {start_flag=1;}
       ;

command: resvarcommand
       | addvarcommand
       | positioncommand
       | frequencycommand
       | meancommand
       ;

command_a: itercommand
       | samplecommand
       | outputcommand
       | mapcommand
       | seedcommand
       | tlocicommand
       | setcommand
       | ibdcommand
       | groupcommand
       | limitresvarcommand
       | limitaddvarcommand
       | analyzecommand
       ;

samplecommand: SAMPLE FROM INTEGER {sample_from[0]=sample_from[1]=$3;}
       | SAMPLE FROM INTEGER ',' INTEGER {sample_from[0]=$3; sample_from[1]=$5;}
       | START OUTPUT INTEGER {sample_from[0]=sample_from[1]=$3;}
       | START OUTPUT INTEGER ',' INTEGER {sample_from[0]=$3; sample_from[1]=$5;}
       ;

setcommand: SET SYSTEM_VAR INTEGER { syst_var[$2].data.value=$3; syst_var[$2].flag=ST_INTEGER; }
       | SET SYSTEM_VAR REAL { syst_var[$2].data.rvalue=$3; syst_var[$2].flag=ST_REAL; }
       | SET STRING rnum { free($2); yyerror("Unrecognized system variable"); }
       ;

ibdcommand: ESTIMATE IBD STRING ',' ibdlist { set_ibd_list($3,$5); free ($3);}
       | ESTIMATE IBD ibdlist { set_ibd_list(0,$3); }
       ;

analyzecom: STRING {set_analyze($1); free($1);}
       | IBD {set_analyze("IBD");}
       ;

analyzelist: analyzecom
       | analyzelist ',' analyzecom
       ;

analyzecommand: ANALYZE {analysis=0;} analyzelist

seedcommand: SEEDFILE STRING {if(Seedfile) free(Seedfile); Seedfile=$2;}
       | SEEDFILE STRING ',' INTEGER {if(Seedfile) free(Seedfile); Seedfile=$2; if($4) ranseed_set|=1; else ranseed_set&=~1;}
       | SEED SFILE STRING {if(Seedfile) free(Seedfile); Seedfile=$3;}
       | SEED SFILE STRING ',' INTEGER {if(Seedfile) free(Seedfile); Seedfile=$3; if($5) ranseed_set|=1; else ranseed_set&=~1;}
       | SEED INTEGER {
			 if($2<0) yyerror("Seedvalue out of range");
			 else {
				 init_ranf($2);
				 ranseed_set|=2;
			 }
		 }
       ;
mapcommand: MAP STRING rnum ',' rnum {set_map_range($2,$3,$5,-1); free($2);}
       | MALE MAP STRING rnum ',' rnum {set_map_range($3,$4,$6,X_PAT); free($3); }
       | FEMALE MAP STRING rnum ',' rnum {set_map_range($3,$4,$6,X_MAT); free($3); }
       | TOTAL MAP rnum {set_map_range(0,$3,$3,-1);}
       | TOTAL MAP rnum ',' rnum {set_map_range(0,$3,$5,-2);}
       | TOTAL MALE MAP rnum {set_map_range(0,$4,$4,X_PAT);}
       | TOTAL FEMALE MAP rnum {set_map_range(0,$4,$4,X_MAT);}
       | MAP TOTAL rnum {set_map_range(0,$3,$3,-1);}
       | MAP TOTAL rnum ',' rnum {set_map_range(0,$3,$5,-2);}
       | MALE MAP TOTAL rnum {set_map_range(0,$4,$4,X_PAT);}
       | FEMALE MAP TOTAL rnum {set_map_range(0,$4,$4,X_MAT);}
       ;

tlocicommand: TRAIT LOCI INTEGER {set_tloci(-1,$3);}
       | START TRAIT LOCI INTEGER {set_tloci(-2,$4);}
       | TRAIT LOCI INTEGER ',' INTEGER {set_tloci($3,$5);}
       | TRAIT LOCI MEAN rnum {tloci_mean=$4; tloci_mean_set=1;}
       ;

itercommand: ITERATIONS INTEGER { num_iter=$2; }

outputcommand: OUTPUT FREQUENCY INTEGER {sample_freq[0]=sample_freq[1]=$3;}
       | OUTPUT FREQUENCY INTEGER ',' INTEGER {sample_freq[0]=$3; sample_freq[1]=$5;}
       | OUTPUT FREQUENCY STRING {if(Freqfile) free(Freqfile); Freqfile=$3;}
       | SAMPLE FREQUENCY INTEGER {sample_freq[0]=sample_freq[1]=$3;}
       | SAMPLE FREQUENCY INTEGER ',' INTEGER {sample_freq[0]=$3; sample_freq[1]=$5;}
       | OUTPUT PHENO STRING {Output_Phen=$3; }
       | OUTPUT GENO STRING {set_output_gen($3,0);} 
       | OUTPUT GENO STRING ',' STRING {set_output_gen($3,$5); free($5); }
       | OUTPUT lkvarlist
       | OUTPUT TYPE INTEGER {output_type=$3;}
       | OUTPUT SFILE STRING {if(Outputfile) free(Outputfile); Outputfile=$3;}
       | DUMP SFILE STRING {if(Dumpfile) free(Dumpfile); Dumpfile=$3;}
       | DUMP FREQUENCY INTEGER {dump_freq=$3;}
		 | OUTPUT HAPLO STRING {output_haplo=1;if(Haplofile) free(Haplofile); Haplofile=$3;}
		 | OUTPUT HAPLO {output_haplo=1;}
		 ;

resvarcommand: RESIDUAL VARIANCE rnum { if(!check_variance($3)) {res_var_set=start_flag; residual_var=$3;} }
          | RESIDUAL VARIANCE NORMAL { use_student_t=0; }
          | RESIDUAL VARIANCE STUDENT_T { use_student_t=1; }
          ;
			 
limitresvarcommand: RESIDUAL VARIANCE LIMIT rnum { if(!check_variance($4)) residual_var_limit=$4; }
          | LIMIT RESIDUAL VARIANCE rnum { if(!check_variance($4)) residual_var_limit=$4; }
          ;

addvarcommand: GENETIC VARIANCE rnum { if(!check_variance($3)) {add_var_set=start_flag; additive_var=$3;} }

limitaddvarcommand: GENETIC VARIANCE LIMIT rnum { if(!check_variance($4)) additive_var_limit=$4; }
          | LIMIT GENETIC VARIANCE rnum { if(!check_variance($4)) additive_var_limit=$4; }
          ;

meancommand: MEAN rnum { grand_mean_set=start_flag; grand_mean=$2; }

positioncommand: POSITION lkvar rnum { set_position($2,$3,$3); }
        | POSITION lkvar rnum ',' rnum { sex_map=1; set_position($2,$3,$5); }
        ;

frequencycommand: FREQUENCY {c_flag=0;} lkmarker freqlist
		  | FREQUENCY {c_flag=1;} COUNTS lkmarker freqlist
        | COUNTS {c_flag=1;} lkmarker freqlist
        ;

lkvar: STRING { $$=find_var($1,0,0); free($1);}
       | STRING '(' INTEGER ')' { $$=find_var($1,$3,1); free($1);}
       ;

lkvarlist: lkvar { if($1) {set_output($1); free($1);} }
       | lkvarlist ',' lkvar { if($3) {set_output($3); free($3);} }
       ;

lkmarker: lkvar { freq_marker=check_marker($1); }

groupcommand: GROUP ORDER {group_ptr=0;} grouplist {if(group_var && group_ptr<group_var->n_levels) print_scan_err("Line %d: Too few groups in order statement\n",lineno1);}
       ;

group: INTEGER { $$=find_group(yytext,$1,1); }
       | REAL { $$=find_group(yytext,0,0); }
       | STRING { $$=find_group($1,0,0); free($1);}
       ;

grouplist: group {set_group_order($1);}
       | grouplist ',' group {set_group_order($3);}
       ;

allele: INTEGER { $$=find_allele(yytext,freq_marker,$1,1); }
       | REAL { $$=find_allele(yytext,freq_marker,0,0); }
       | STRING { $$=find_allele($1,freq_marker,0,0); free($1); }
       ;

freqlist: allele ',' {group_counter=0; freq_allele=$1;} freqlist1 {if(freq_marker && group_var && group_counter<group_var->n_levels) print_scan_err("Line %d: Too few frequencies specified\n",lineno);}
        | freqlist allele ',' {group_counter=0; freq_allele=$2;} freqlist1 {if(freq_marker && group_var && group_counter<group_var->n_levels) print_scan_err("Line %d: Too few frequencies specified\n",lineno);}
        ;

freqlist1: rnum {set_freq(freq_marker,$1,freq_allele);}
        | '*' {group_counter++;}
        | freqlist1 ',' rnum {set_freq(freq_marker,$3,freq_allele);}
        | freqlist1 ',' '*' {group_counter++;}
        ;

ibdlist: rnum { $$=add_ibd_list($1,0); }
        | ibdlist ',' rnum { $$=add_ibd_list($3,$1); }
        ;

rnum: REAL
    | INTEGER { $$=(double)$1; }
    ;

%%

static void set_analyze(char *p)
{
	int i;
	char *com[]={"AFFECTED","NULL","IBD","IBS",0};
		
	if(p) {
		i=0;
		while(com[i]) {
			if(!strcasecmp(com[i],p)) break;
			i++;
		}
		if(com[i]) analysis|=(1<<i);
		else yyerror("Invalid parameter to analyze statement");
	}
}

static void set_group_order( int gp)
{
	int i;
	
	if(!group_var || gp<0) return;
	for(i=0;i<group_ptr;i++) if(group_order[i]==gp)	{
		yyerror("Group repeated in order statement");
		return;
	}
	if(group_ptr>=group_var->n_levels) yyerror("Too many groups - internal error?");
	else group_order[group_ptr++]=gp;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "set_output_gen"
static void set_output_gen(char *file,char *link)
{
	int i;
	struct output_gen *p;

	if(link)	{
		for(i=0;i<n_links;i++) if(!strcasecmp(link,linkage[i].name)) break;
		if(i==n_links) return;
		i++;
	} else i=0;
	p=Output_Gen;
	if(!(Output_Gen=malloc(sizeof(struct output_gen)))) ABT_FUNC(MMsg);
	Output_Gen->next=p;
	Output_Gen->file=file;
	Output_Gen->link_group=i;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "add_ibd_list"
static struct IBD_List *add_ibd_list(double x,struct IBD_List *p)
{
	if(!p) {
		if(!(p=malloc(sizeof(struct IBD_List)))) ABT_FUNC(MMsg);
		p->idx=0;
		p->size=32;
		if(!(p->pos=malloc(sizeof(double)*p->size))) ABT_FUNC(MMsg);
	}
	if(p->size==p->idx) {
		p->size*=2;
		if(!(p->pos=realloc(p->pos,sizeof(double)*p->size))) ABT_FUNC(MMsg);
	}
	p->pos[p->idx++]=x;
	return p;
}

static void set_ibd_list(char *name,struct IBD_List *p)
{
	int i;
	
	if(name)	{
		for(i=0;i<n_links;i++) if(!strcasecmp(name,linkage[i].name)) break;
		if(i==n_links) {
			fprintf(stderr,"Warning: linkage group %s not found; IBD positions ignored\n",name);
			i= -1;
		}
	} else {
		if(n_links==1) i=0;
		else {
			i= -1;
			fprintf(stderr,"Warning: linkage group not specified when number of linkage groups >1; IBD positions ignored\n");
		}
	}
	if(i<0) {
		free(p->pos);
		free(p);
	} else {
		if(linkage[i].ibd_list)	{
			fprintf(stderr,"Warning: overwriting list of IBD positions for linkage group %s\n",linkage[i].name);
			free(linkage[i].ibd_list->pos);
			free(linkage[i].ibd_list);
		}
		linkage[i].ibd_list=p;
	}
}

static void set_tloci(int a,int b)
{
	if(a<0) {
		if(a== -1) min_tloci=max_tloci=b;
		else start_tloci=b;
	} else if(a<b) {
		min_tloci=a;
		max_tloci=b;
	} else {
		min_tloci=b;
		max_tloci=a;
	}
}

static struct Marker *check_marker(struct lk_variable *lkvar)
{
	struct Marker *mk=0;
	
	if(!lkvar) return 0;
	if(lkvar->type!=LK_TYPE_MARKER)
		yyerror("Attempting to set frequency of a non-marker");
	else mk=lkvar->var.marker;
	free(lkvar);
	return mk;
}

static void set_output(struct lk_variable *lkvar)
{
	int i,j,type;
	struct Variable *var;
	
	if(!lkvar) return;

	for(i=0;i<model.n_terms;i++)
	{
		type=model.term[i].vars[0].type;
		j=model.term[i].vars[0].var_index;
		if(type&ST_MARKER)
		{
			if(lkvar->type==LK_TYPE_MARKER && lkvar->var.marker==marker+j)
			{
				model.term[i].out_flag=1;
				return;
			}
		}
		else if(type&ST_TRAITLOCUS) continue;
		else
		{
			if(type&ST_CONSTANT) var=id_variable+j;
			else var=nonid_variable+j;
			if((lkvar->type==LK_TYPE_IDVAR || lkvar->type==LK_TYPE_NONIDVAR) && lkvar->var.var==var)
			{
				model.term[i].out_flag=1;
				return;
			}
		}
	}
}

static void set_map_range( char *name,double r1,double r2, int flag)
{
	int i;
	double t;
	static char *sexstr[2]={"female","male"};
	
	if(flag!= -1) sex_map=1;
	
	if(name)	{
		for(i=0;i<n_links;i++) if(!strcasecmp(name,linkage[i].name)) {
			if(r2<r1) {
				t=r2;
				r2=r1;
				r1=t;
			}
			if(flag== -1) {
				linkage[i].r1[0]=linkage[i].r1[1]=r1;
				linkage[i].r2[0]=linkage[i].r2[1]=r2;
				linkage[i].range_set[0]=linkage[i].range_set[1]=1;
				printf("Map range for linkage group '%s' set to %g-%gcM\n",name,r1,r2);
			} else {
				linkage[i].r1[flag]=r1;
				linkage[i].r2[flag]=r2;
				linkage[i].range_set[flag]=1;
				printf("Map range (%s) for linkage group '%s' set to %g-%gcM\n",sexstr[flag],name,r1,r2);
			}
			break;
		}
	} else {
		if(flag<0) {
			total_maplength[X_PAT]=r1;
			total_maplength[X_MAT]=r2;
			printf("Total (genome) map length set to (%g,%g)cM\n",r1,r2);
		} else {
			total_maplength[flag]=r1;
			printf("Total (genome) %s map length set to %gcM\n",sexstr[flag],r1);
		}
	}
}

static int find_group( char *p,int gp, int flag)
{
	int i,j;
	char *s;
	
	if(!group_var) return -1;
	j=group_var->n_levels;
	
	if(group_var->rec_flag==ST_STRING) {
		for(i=0;i<j;i++) if(!(strcasecmp(p,group_var->recode[i].string))) return i;
	} else {
		if(!flag) {
			gp=strtol(p,&s,10);
			if(!(*s)) for(i=0;i<j;i++) if(gp==group_var->recode[i].value) return i;
		} else for(i=0;i<j;i++) if(gp==group_var->recode[i].value) return i;
	}
	yyerror("Group not found\n");
	return -1;
}

static int find_allele( char *p, struct Marker *mk,int all, int flag)
{
	int i,j;
	char *s;
	
	if(!mk) return -1;
	j=extra_allele_flag?mk->n_alleles:mk->n_alleles-1;
	
	if(mk->rec_flag==ST_STRING) {
		for(i=0;i<j;i++) if(!(strcmp(p,mk->recode[i].string))) return i;
	} else {
		if(!flag) {
			all=strtol(p,&s,10);
			if(!(*s)) for(i=0;i<j;i++) if(all==mk->recode[i].value) return i;
		} else for(i=0;i<j;i++) if(all==mk->recode[i].value) return i;
	}
	return -1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_var"
static struct lk_variable *find_var( char *p, int idx, int flag)
{
	int i,j=0;
	struct lk_variable *lkv=0;
	
	if(flag) if(idx<1) {
		yyerror("Illegal array index");
		return 0;
	}
	for(i=0;i<n_markers;i++) if(!strcasecmp(marker[i].name,p)) {
		if(flag && idx!=marker[i].index) continue;
		j=LK_TYPE_MARKER;
		break;
	}
	if(!j) for(i=0;i<n_id_records;i++) if(!strcasecmp(id_variable[i].name,p)) {
		if(flag && idx!=id_variable[i].index) continue;
		j=LK_TYPE_IDVAR;
		break;
	}
	if(!j) for(i=0;i<n_nonid_records;i++) if(!strcasecmp(nonid_variable[i].name,p)) {
		if(flag && idx!=nonid_variable[i].index) continue;
		j=LK_TYPE_NONIDVAR;
		break;
	}
	if(!j && !flag) for(i=0;i<n_links;i++) if(!strcasecmp(linkage[i].name,p)) {
		j=LK_TYPE_LINK;
		break;
	}
	if(j) {
		if(!(lkv=malloc(sizeof(struct lk_variable)))) ABT_FUNC(MMsg);
		lkv->type=j;
		switch(j) {
		 case LK_TYPE_MARKER:
			lkv->var.marker=marker+i;
			break;
		 case LK_TYPE_LINK:
			lkv->var.link=linkage+i;
			break;
		 case LK_TYPE_IDVAR:
			lkv->var.var=id_variable+i;
			break;
		 case LK_TYPE_NONIDVAR:
			lkv->var.var=nonid_variable+i;
			break;
		}
	}
	return lkv;
}

static void set_freq(struct Marker *mk, double freq,int allele)
{
	static int fg,fg1;
	int i=0;
	
	group_counter++;
	if(mk) {
		if(group_var) {
			if(!group_ptr)	{
				if(!fg) yyerror("Genetic group order not set");
				fg=1;
				return;
			}
			if(group_counter>group_var->n_levels) {
				if(!fg1) yyerror("Too many frequencies specified (only 1 per genetic group)");
				fg1=1;
				return;
			}
			i=group_order[group_counter-1];
			if(i<0) return;
		}
		if(freq<0.0) {
			yyerror("Invalid (negative) frequency");
			return;
		}
		if(allele>=0 && freq==0.0)	{
			yyerror("Can not set frequency of observed allele to zero\n");
			return;
		}
		if(allele>=0) {
			mk->freq[i][allele]=freq;
		} else {
			if(extra_allele_flag) {
				yyerror("Cannot set frequency of unknown allele\n");
				return;
			}
			allele=mk->n_alleles-1;
			mk->freq[i][allele]+=freq;
		}
		mk->freq_set[i][allele]=start_flag;
		mk->count_flag[i]=c_flag;
	}
	fg1=0;
}

static void set_position(struct lk_variable *lkvar, double pos1, double pos2)
{
	if(lkvar) {
		if(lkvar->type!=LK_TYPE_MARKER) {
			yyerror("Attempting to set position of a non-marker");
			return;
		}
		lkvar->var.marker->pos[X_PAT]=pos1;
		lkvar->var.marker->pos[X_MAT]=pos2;
		lkvar->var.marker->pos_set=start_flag;
		free(lkvar);
	}
}

static int check_variance( double v)
{
	if(v<=0.0) {
		yyerror("Variance must be positive");
		return 1;
	}
	return 0;
}

void print_scan_err(char *fmt, ...)
{
	va_list args;
	
	va_start(args,fmt);
	vfprintf(stderr,fmt,args);
	va_end(args);
	if((++scan_error_n)>=max_scan_errors) abt(__FILE__,__LINE__,"Too many errors - aborting\n");
}

void print_scan_warn(char *fmt, ...)
{
	va_list args;
	
	if(scan_warn_n<max_scan_warnings)
	{
		va_start(args,fmt);
		vfprintf(stderr,fmt,args);
		va_end(args);
	}
	scan_warn_n++;
}

void yyerror(char *s)
{
     int i;
	
	print_scan_err("Line %d: %s\n%s\n",lineno,s,linebuf);
	if(scan_error_n<=max_scan_errors)
	{
		for(i=1;i<tokenpos;i++) putc('-',stderr);
		fputs("^\n",stderr);
	}
}

int yywrap(void)
{
	return 1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "ReadParam"
int ReadParam(FILE *fptr)
{
	int i,j;
	void yy_cleanup(void);
	
	for(i=0;i<NUM_SYSTEM_VAR;i++) syst_var[i].flag=0;
	for(i=0;i<n_id_records;i++) if(id_variable[i].type&ST_GROUP) {
		group_var=id_variable+i;
		j=group_var->n_levels;
		if(!(group_order=malloc(sizeof(int)*j))) ABT_FUNC(MMsg);
		break;
	}
	yyin=fptr;
	yyparse();
	yy_cleanup();
	if(group_order) free(group_order);
	if(start_tloci<0) start_tloci=min_tloci;
	else if(start_tloci<min_tloci || start_tloci>max_tloci) {
		fprintf(stderr,"ReadParam(): Starting no. trait loci (%d) is outside set range (%d-%d)\n",start_tloci,min_tloci,max_tloci);
		scan_error_n++;
	}
	if(scan_error_n) return 1;
	return 0;
}
