/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       July 1997                                          *
 *                                                                          *
 * seg_pen.c:                                                               *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"
#include "loki.h"
#include "loki_peel.h"

static double **seg_freq,*seg_ll[2],seg_log2;
static int *seg_obslist,*seg_alls;

double addlog(double x1,double x2)
{
	double yy, r;

	yy=.5*(x1+x2);
	r=(log(cosh(x1-yy))+yy+seg_log2);
	return r;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "setup_groups"
static void setup_groups(void)
{
	int i,i1,j,k,comp,locus,nfd=0;
	int **temp_p,*temp_p1;
	
	for(locus=0;locus<n_markers;locus++) {
		for(i=comp=0;comp<n_comp;comp++) {
			for(j=0;j<comp_size[comp];j++) {
				i1=i+j;
				k=id_array[i1].sire;
				if(!k || id_array[k-1].pruned_flag[locus]) nfd+=2;
			}
			i+=j;
		}
	}
	if(!nfd) return;
	if(!(temp_p=malloc(sizeof(void *)*n_markers*n_comp))) ABT_FUNC(MMsg);
	RemBlock=AddRemem(temp_p,RemBlock);
	if(!(temp_p1=malloc(sizeof(int)*nfd))) ABT_FUNC(MMsg);
	RemBlock=AddRemem(temp_p1,RemBlock);
	for(locus=0;locus<n_markers;locus++) {
		marker[locus].group=temp_p;
		for(i=comp=0;comp<n_comp;comp++)	{
			temp_p[comp]=temp_p1;
			for(nfd=j=0;j<comp_size[comp];j++) {
				i1=i+j;
				k=id_array[i1].sire;
				if(!k || id_array[k-1].pruned_flag[locus]) {
					temp_p1[nfd++]=id_array[i1].group;
					temp_p1[nfd++]=id_array[i1].group;
					if(!id_array[i1].group) {
						ABT_FUNC("Internal error: null group\n");
					}
				}
			}
			i+=j;
			temp_p1+=nfd;
		}
		temp_p+=n_comp;
	}
}

static void change_genes(const int i,const int par,const int par_flag,const int locus,int **genes,signed char **seg,int *fg)
{
	int g=0,j,k=0,kid;
	signed char s;
	
	s=seg[par_flag][i];
	if(s<0) ABT_FUNC("Internal error - illegal seg.\n");
	if(!*fg && id_array[i].haplo[0][locus]) {
		g=genes[par_flag][i];
		k=1;
	}
	genes[par_flag][i]=genes[s][par];
	if(k && genes[par_flag][i]!=g) *fg=1;
	k=id_array[i].sex==1?X_PAT:X_MAT;
	for(j=0;j<id_array[i].nkids;j++) {
		kid=id_array[i].kids[j];
		if(seg[k][kid]==par_flag) change_genes(kid,i,k,locus,genes,seg,fg);
	}
}

int pass_founder_genes1(const int locus,const int i,const int par_flag)
{
	int **genes,par,fg=0;
	signed char **seg;

	genes=marker[locus].genes;
	seg=marker[locus].seg;
	par=(par_flag==X_MAT)?id_array[i].dam:id_array[i].sire;
	change_genes(i,par-1,par_flag,locus,genes,seg,&fg);
	return fg;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "pass_founder_genes"
int pass_founder_genes(const int locus,const int comp)
{
	int i,i1,j,k,s,**genes,g[2],fg=0;
	signed char **seg;

 	if(marker[locus].n_all1[comp]<2) return fg; 
	genes=marker[locus].genes;
	seg=marker[locus].seg;
	i=comp_start[comp];
	for(i1=j=0;i1<comp_size[comp];i1++,i++) {
		s=seg[X_PAT][i];
		if(s<-1) continue;
		if(!fg && id_array[i].haplo[0][locus]) {
			g[X_PAT]=genes[X_PAT][i];
			g[X_MAT]=genes[X_MAT][i];
			k=1;
		} else k=0;
		if(s>=0)	{
			genes[X_PAT][i]=genes[s][id_array[i].sire-1];
		} else genes[X_PAT][i]= ++j;
		s=seg[X_MAT][i];
		if(s>=0) {
			genes[X_MAT][i]=genes[s][id_array[i].dam-1];
		} else genes[X_MAT][i]= ++j;
		if(k) {
			for(s=0;s<2;s++) if(genes[s][i]!=g[s]) {
				fg=1;
				break;
			}
		}
	}
	if(j>comp_ngenes[comp]) {
		ABT_FUNC("Internal error - this should not happen\n");
	}
	return fg;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "seg_alloc"
void seg_alloc(void)
{
	int i,j=0,k=0,k1;

	setup_groups();
	for(i=0;i<n_markers;i++) {
		j=marker[i].n_alleles;
		if(j>k) k=j;
	}
	if(!j) return;
	if(!(seg_freq=malloc(sizeof(void *)*n_genetic_groups))) ABT_FUNC(MMsg);
	if(!(seg_freq[0]=malloc(sizeof(double)*k*n_genetic_groups))) ABT_FUNC(MMsg);
	for(i=1;i<n_genetic_groups;i++) seg_freq[i]=seg_freq[i-1]+k;
	if(!(seg_alls=malloc(sizeof(int)*k))) ABT_FUNC(MMsg);
	for(k=k1=i=0;i<n_comp;i++) {
		j=comp_size[i];
		if(j>k) k=j;
		j=comp_ngenes[i];
		if(j>k1) k1=j;
	}
	if(!(seg_obslist=malloc(sizeof(int)*(k1*6+k)))) ABT_FUNC(MMsg);
	if(!(seg_ll[0]=malloc(sizeof(double)*k1*2))) ABT_FUNC(MMsg);
	seg_ll[1]=seg_ll[0]+k1;
	seg_log2=log(2.0);
}

void seg_dealloc(void)
{
	if(seg_freq) {
		if(seg_freq[0]) free(seg_freq[0]);
		free(seg_freq);
	}
	if(seg_alls) free(seg_alls);
	if(seg_obslist) free(seg_obslist);
	if(seg_ll[0]) free(seg_ll[0]);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "seg_pen"
double seg_pen(int locus,int comp,int *err,int flag)
{ 
	int i,ii,i1,j,k,k1,s,t,a1,a2,g1,g2,o1,o2;
	int ngenes,*group,*code,*alleles[2],*tr,**genes;
	int nobs,*groupflag,nn_all,n_all,idd,ids;
	signed char **seg;
	double l,z,z1;

	n_all=marker[locus].n_all1[comp];
	if(n_all<2) return 0.0;
	*err=0;
	genes=marker[locus].genes;
	ngenes=comp_ngenes[comp];
	memset(seg_obslist,0,sizeof(int)*(comp_size[comp]+6*ngenes));
	nobs=0;
	group=seg_obslist+comp_size[comp];
	code=group+ngenes;
	tr=code+ngenes;
	i=comp_start[comp];
	/* Group together genes which occur in observed individuals */
	for(j=i1=0;i1<comp_size[comp];i1++,i++) {
		id_array[i].allele[0]=id_array[i].allele[1]=0;
		if(id_array[i].haplo[0][locus]) id_array[i].flag=1;
		else {
			id_array[i].flag=0;
			continue;
		}
		seg_obslist[nobs++]=i;
		a1=genes[X_MAT][i];
		a2=genes[X_PAT][i];
		g1=group[a1-1];
		g2=group[a2-1];
		if(g1) {
			if(g2) {
				g1=tr[g1-1];
				g2=tr[g2-1];
				if(g1<g2) {
					for(k=g1+1;k<j;k++) if(tr[k]==g2) tr[k]=g1;
				} else if(g2<g1) {
					for(k=g2+1;k<j;k++) if(tr[k]==g1) tr[k]=g2;
				}
			} else group[a2-1]=g1;
		} else if(g2) group[a1-1]=g2;
		else {
			tr[j]=j;
			group[a1-1]=group[a2-1]= ++j;
		}
	}
	/* Recode groups */
	for(i=j=0;i<ngenes;i++) {
		s=group[i];
		if(s)	{
			s=tr[s-1];
			if(!code[s]) code[s]= ++j;
			group[i]=code[s];
		}
	}
	/* alleles[0][i] and alleles [1][i] will contain the 2 possible allele
	 * types for founder gene i */
	alleles[0]=tr+ngenes;
	alleles[1]=alleles[0]+ngenes;
	groupflag=alleles[1]+ngenes;
	/* Sort out possible allele allocations for groups */
	while(nobs)	{
		for(ii=0;ii<nobs;ii++) {
			i=seg_obslist[ii];
			o1=id_array[i].haplo[X_PAT][locus];
			o2=id_array[i].haplo[X_MAT][locus];
			a1=genes[X_PAT][i]-1;
			a2=genes[X_MAT][i]-1;
			g1=group[a1]-1;
			if(g1==2) {
				g1=2;
			}
			if(code[a1]<0)	{
				k=0;
				if(code[a2]<0)	{
					if(!(groupflag[g1]&1)) {
						if((alleles[0][a1]==o1 && alleles[0][a2]==o2) ||
						   (alleles[0][a1]==o2 && alleles[0][a2]==o1)) k++;
						else groupflag[g1]|=1;
					}
					if(!(groupflag[g1]&2)) {
						if((alleles[1][a1]==o1 && alleles[1][a2]==o2) ||
							(alleles[1][a1]==o2 && alleles[1][a2]==o1)) k++;
						else groupflag[g1]|=2;
					}
					if(!k) {
						if(flag==1) {
							printf("%d %d %d\n",i,o1,o2);
						}
						*err= -1;
						return 0.0;
					}
				} else {	
					for(k1=t=0;t<2;t++) if(!(groupflag[g1]&(1<<t)))	{
						k=0;
						if(alleles[t][a1]==o1) {
							alleles[t][a2]=o2;
							k++;
						} else if(alleles[t][a1]==o2)	{
							alleles[t][a2]=o1;
							k++;
						}
						if(!k) groupflag[g1]|=(1<<t);
						else k1+=k;
					}
					if(!k1) {
						*err = -2;
						return 0.0;
					}
					code[a2]= -1;
				}
			} else if(code[a2]<0) {
				for(k1=t=0;t<2;t++) if(!(groupflag[g1]&(1<<t))) {
					k=0;
					if(alleles[t][a2]==o1) {
						alleles[t][a1]=o2;
						k++;
					} else if(alleles[t][a2]==o2)	{
						alleles[t][a1]=o1;
						k++;
					}
					if(!k) groupflag[g1]|=(1<<t);
					else k1+=k;
				}
				if(!k1) {
					*err= -3;
					return 0.0;
				}
				code[a1]= -1;
			} else {
				/* Check if another gene in the group is present */
				/* If so then skip this observation 'till later */
				if(groupflag[g1]) continue;
				/* First genes in group */
				code[a1]=code[a2]= -1;
				alleles[0][a1]=o1;
				alleles[0][a2]=o2;
				if(o1!=o2) {
					alleles[1][a1]=o2;
					alleles[1][a2]=o1;
					groupflag[g1]=4;
				} else groupflag[g1]=6;
			}
			seg_obslist[ii]=seg_obslist[--nobs];
		}
	}
	nn_all=marker[locus].n_alleles;
	for(i=0;i<nn_all;i++) {
		for(k=0;k<n_genetic_groups;k++) seg_freq[k][i]=0.0;
		seg_alls[i]=n_all-1;
	}
	for(i=0;i<nn_all;i++) if((k=allele_trans[locus][comp][i])>=0) seg_alls[k]=i;
	for(i=0;i<nn_all;i++) for(k=0;k<n_genetic_groups;k++)
	  seg_freq[k][seg_alls[i]]+=marker[locus].freq[k][i];
	for(g1=0;g1<j;g1++) seg_ll[0][g1]=seg_ll[1][g1]=0.0;
	l=0.0;
	if(flag) {
		for(i=0;i<ngenes;i++) {
			code[i]=0;
			g1=group[i]-1;
			if(g1<0) continue;
			k=marker[locus].group[comp][i]-1;
			if(!(groupflag[g1]&3)) {
				seg_ll[0][g1]+=log(seg_freq[k][alleles[0][i]-1]);
				seg_ll[1][g1]+=log(seg_freq[k][alleles[1][i]-1]);
			}
		}
		for(g1=0;g1<j;g1++) {
			if(!(groupflag[g1]&3)) {
				z=exp(seg_ll[0][g1]);
				z1=exp(seg_ll[1][g1]);
			   k=(ranf()*(z+z1)<z)?2:1;
				groupflag[g1]|=k;
			}
		}
		for(i=0;i<ngenes;i++) {
			code[i]=0;
			g1=group[i]-1;
			if(g1<0) continue;
			if(groupflag[g1]&1) code[i]=alleles[1][i];
			else if(groupflag[g1]&2) code[i]=alleles[0][i];
			else ABT_FUNC("Internal error\n");
		}
		i=comp_start[comp];
		seg=marker[locus].seg;
		for(i1=0;i1<comp_size[comp];i1++,i++) {
			if(id_array[i].pruned_flag[locus]) continue;
			a1=genes[X_MAT][i]-1;
			a2=genes[X_PAT][i]-1;
			id_array[i].allele[X_MAT]=code[a1];
			id_array[i].allele[X_PAT]=code[a2];
			if(code[a1]) {
				idd=id_array[i].dam;
				if(idd && !id_array[idd-1].pruned_flag[locus]) {
					k=code[a1];
					if(k!=id_array[idd-1].allele[seg[X_MAT][i]]) {
						ABT_FUNC("Internal error - bad sampled allele\n");
					}
					if(id_array[idd-1].allele[X_MAT]==id_array[idd-1].allele[X_PAT]) {
						seg[X_MAT][i]=-2;
					}
				}
			}
			if(code[a2]) {
				ids=id_array[i].sire;
				if(ids && !id_array[ids-1].pruned_flag[locus]) {
					k=code[a2];
					if(k!=id_array[ids-1].allele[seg[X_PAT][i]]) ABT_FUNC("Internal error - bad sampled allele\n");
					if(id_array[ids-1].allele[X_MAT]==id_array[ids-1].allele[X_PAT]) {
						seg[X_PAT][i]=-2;
					}
				}
			}
		}
	} else {
		for(i=0;i<ngenes;i++) {
			g1=group[i]-1;
			if(g1<0) continue;
			k=marker[locus].group[comp][i]-1;
			if(groupflag[g1]&1) l+=log(seg_freq[k][alleles[1][i]-1]);
			else if(groupflag[g1]&2) l+=log(seg_freq[k][alleles[0][i]-1]);
			else {
				seg_ll[0][g1]+=log(seg_freq[k][alleles[0][i]-1]);
				seg_ll[1][g1]+=log(seg_freq[k][alleles[1][i]-1]);
			}
		}
		for(g1=0;g1<j;g1++) {
			if(!(groupflag[g1]&3)) {
				if(seg_ll[0][g1]==seg_ll[1][g1]) l+=seg_ll[0][g1]+seg_log2;
				else l+=addlog(seg_ll[0][g1],seg_ll[1][g1]);
			}
		}
	}
	return l;
}
